# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from abc import ABCMeta, abstractmethod
import numpy as np
from cubical.flagging import FL
from cubical.machines.interval_gain_machine import PerIntervalGains

class ParameterisedGains(PerIntervalGains):
    """
    This is a base class for all gain solution machines that use full-resolution
    gain arrays derived from some other parameters.
    """

    def init_gains(self):
        """
        Construct gain and flag arrays at full-resolution. Parameterisation of gains may not be at
        full resolution.
        """
        self.gain_intervals = 1, 1
        self.gain_grid = self.data_grid
        self.gain_shape = [self.n_dir,self.n_tim,self.n_fre,self.n_ant,self.n_cor,self.n_cor]
        self.gains = np.empty(self.gain_shape, dtype=self.dtype)
        self.gains[:] = np.eye(self.n_cor)
        self.gflags = np.zeros(self.gain_shape[:-2],FL.dtype)

        # Function used to unpack the gains or flags into full time/freq resolution.
        self._gainres_to_fullres  = self.copy_or_identity

        # function used to unpack interval resolution to gain resolution.
        self._interval_to_gainres = self.unpack_intervals




'''
DDFacet, a facet-based radio imaging package
Copyright (C) 2013-2016  Cyril Tasse, l'Observatoire de Paris,
SKA South Africa, Rhodes University

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

This file has been ported to CubiCal by the original authors, with minor modifications.
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DDFacet.compatibility import range

import numpy
import os
import os.path
import sys

from cubical.tools import logger, ModColor
log = logger.getLogger("FITSBeamInterpolator")

import pyrap.tables

import numpy as np


dm = pyrap.measures.measures()
dq = pyrap.quanta

# This a list of the Stokes enums (as defined in casacore header measures/Stokes.h)
# These are referenced by the CORR_TYPE column of the MS POLARIZATION subtable.
# E.g. 5,6,7,8 corresponds to RR,RL,LR,LL
MS_STOKES_ENUMS = [
    "Undefined", "I", "Q", "U", "V", "RR", "RL", "LR", "LL", "XX", "XY", "YX", "YY", "RX", "RY", "LX", "LY", "XR", "XL", "YR", "YL", "PP", "PQ", "QP", "QQ", "RCircular", "LCircular", "Linear", "Ptotal", "Plinear", "PFtotal", "PFlinear", "Pangle"
  ];
# set of circular correlations
CIRCULAR_CORRS = set(["RR", "RL", "LR", "LL"]);
# set of linear correlations
LINEAR_CORRS = set(["XX", "XY", "YX", "YY"]);


class FITSBeamInterpolator (object):
    def __init__ (self, 
                  field_centre,
                  chunk_ctr_frequencies,
                  chunk_channel_widths,
                  correlation_ids,
                  station_positions,
                  opts):
        """
            field_centre: radian ra dec coordinate of the original field phase centre
            chunk_ctr_frequencies: center channel frequencies (Hz) of chunk to predict E terms for
            chunk_channel_widths: channel widths of chunk channels
            correlation_ids: hand labels per correlation term as defined in casacore Stokes.h
            station_positions: ECEF coordinates for stations (na, 3) array
            opts: beam opts as defined in Parset
        """
        # filename is potentially a list (frequencies will be matched)
        self.beamsets = opts["FITSFile"]
        if type(self.beamsets) is not list:
            self.beamsets = self.beamsets.split(',')
        self.pa_inc = opts["FITSParAngleIncDeg"]
        self.time_inc = opts["DtBeamMin"]
        self.nchan = opts["NBand"]
        self.feedangle = opts["FeedAngle"]
        self.applyrotation = False # Handled by Cubical ParallacticMachine after visibilities have been predicted
                                   # ie P . degrid(E, X, E^H) . P^H
        self.applyantidiagonal = opts["FlipVisibilityHands"]
        self.station_positions = station_positions

        # make masure for zenith
        self.zenith = dm.direction('AZEL','0deg','90deg')
        # make position measure from antenna 0
        # NB: in the future we may want to treat position of each antenna separately. For
        # a large enough array, the PA w.r.t. each antenna may change! But for now, use
        # the PA of the first antenna for all calculations
        self.pos0 = dm.position('itrf',*[ dq.quantity(x,'m') for x in self.station_positions[0] ]) 

        # make direction measure from field centre
        ra, dec = tuple(field_centre)
        self.field_centre_radec = tuple([ra, dec])
        self.field_centre = dm.direction('J2000',dq.quantity(ra,"rad"),dq.quantity(dec,"rad"))

        # get channel frequencies from MS
        self.freqs = chunk_ctr_frequencies.ravel()
        if not self.nchan:
            self.nchan = len(self.freqs)
        else:
            cw = chunk_channel_widths.ravel()          
            fq = np.linspace(self.freqs[0]-cw[0]/2, self.freqs[-1]+cw[-1]/2, self.nchan+1)
            self.freqs = (fq[:-1] + fq[1:])/2

        feed = opts["FITSFeed"]
        if feed:
            if len(feed) != 2:
                raise ValueError("FITSFeed parameter must be two characters (e.g. 'xy')")
            feed = feed.lower()
            if "x" in feed:
                self.feedbasis = "linear"
            else:
                self.feedbasis = "circular"
            self.corrs = [ a+b for a in feed for b in feed ]
            print("polarization basis specified by FITSFeed parameter: %s"%" ".join(self.corrs), file=log)
        else:
            # NB: need to check correlation names better. This assumes four correlations in that order!
            if set(correlation_ids) <= set([MS_STOKES_ENUMS.index(c) for c in LINEAR_CORRS]):
                self.corrs = "xx","xy","yx","yy"
                self.feedbasis = "linear"
                print("polarization basis is linear (MS corrs: %s)"%" ".join([MS_STOKES_ENUMS[c] for c in correlation_ids]), file=log)
            elif set(correlation_ids) <= set([MS_STOKES_ENUMS.index(c) for c in CIRCULAR_CORRS]):
                self.corrs = "rr","rl","lr","ll"
                self.feedbasis = "circular"
                print("polarization basis is circular (MS corrs: %s)"%" ".join([MS_STOKES_ENUMS[c] for c in correlation_ids]), file=log)
            else:
                raise RuntimeError("FITS interpolator currently only support linear or circular correlations.")
                
        if opts["FITSFeedSwap"]:
            print("swapping feeds as per FITSFeedSwap setting", file=log)
            self._feed_swap_map = dict(x="y", y="x", r="l", l="r")
        else:
            self._feed_swap_map = None
        # Following code is nicked from Cattery/Siamese/OMS/pybeams_fits.py
        REALIMAG = dict(re="real",im="imag");

        # get the Cattery: if an explicit path to Cattery set, use this and import Siamese directly
        explicit_cattery = False
        for varname in "CATTERY_PATH","MEQTREES_CATTERY_PATH":
            if varname in os.environ:
                sys.path.append(os.environ[varname])
                explicit_cattery = True

        if explicit_cattery:
            import Siamese.OMS.Utils as Utils
            import Siamese
            import Siamese.OMS.InterpolatedBeams as InterpolatedBeams
            print("explicit Cattery path set: using custom Siamese module from %s"%os.path.dirname(Siamese.__file__), file=log)
        else:
            import Cattery.Siamese.OMS.Utils as Utils
            import Cattery.Siamese as Siamese
            import Cattery.Siamese.OMS.InterpolatedBeams as InterpolatedBeams
            print("using standard Cattery.Siamese module from %s"%os.path.dirname(Siamese.__file__), file=log)

        def make_beam_filename (filename_pattern,corr,reim):
            """Makes beam filename for the given correlation and real/imaginary component (one of "re" or "im")"""
            return Utils.substitute_pattern(filename_pattern,
                      corr=corr.lower(),xy=corr.lower(),CORR=corr.upper(),XY=corr.upper(),
                      reim=reim.lower(),REIM=reim.upper(),ReIm=reim.title(),
                      realimag=REALIMAG[reim].lower(),REALIMAG=REALIMAG[reim].upper(),
                      RealImag=REALIMAG[reim].title());

        self.vbs = {}

        # now, self.beamsets specifies a list of filename patterns. We need to find the one with the closest
        # frequency coverage

        if isinstance(opts["FITSFile"], str) and opts["FITSFile"].upper() == "UNITY":
            self.use_unity_ejones = True
        else:
            self.use_unity_ejones = False
            for corr in self.corrs:
                beamlist = []
                corr1 = "".join([self._feed_swap_map[x] for x in corr])if self._feed_swap_map else corr
                for beamset in self.beamsets:
                    filenames = make_beam_filename(beamset, corr1, 're'), make_beam_filename(beamset, corr1, 'im')
                    # get interpolator from cache, or create object
                    vb = FITSBeamInterpolator._vb_cache.get(filenames)
                    if vb is None:
                        print("loading beam patterns %s %s" % filenames, file=log)
                        FITSBeamInterpolator._vb_cache[filenames] = vb = InterpolatedBeams.LMVoltageBeam(
                            verbose=opts["FITSVerbosity"],
                            l_axis=opts["FITSLAxis"], m_axis=opts["FITSMAxis"]
                        )  # verbose, XY must come from options
                        vb.read(*filenames)
                    else:
                        print("beam patterns %s %s already in memory" % filenames, file=log)
                    # find frequency "distance". If beam frequency range completely overlaps MS frequency range,
                    # this is 0, otherwise a positive number
                    distance = max(vb._freqgrid[0] - self.freqs[0], 0) + \
                               max(self.freqs[-1] - vb._freqgrid[-1], 0)
                    beamlist.append((distance, vb, filenames))
                # select beams with smallest distance
                dist0, vb, filenames = sorted(beamlist)[0]
                if len(beamlist) > 1:
                    if dist0 == 0:
                        print("beam patterns %s %s overlap the frequency coverage" % filenames, file=log)
                    else:
                        print("beam patterns %s %s are closest to the frequency coverage (%.1f MHz max separation)" % (
                                        filenames[0], filenames[1], dist0*1e-6), file=log)
                    print("  MS coverage is %.1f to %.1f GHz, beams are %.1f to %.1f MHz"%(
                        self.freqs[0]*1e-6, self.freqs[-1]*1e-6, vb._freqgrid[0]*1e-6, vb._freqgrid[-1]*1e-6), file=log)
                self.vbs[corr] = vb


    _vb_cache = {}

    def getBeamSampleTimes (self, times, quiet=False):
        """For a given list of timeslots, returns times at which the beam must be sampled"""
        if not quiet:
            print("computing beam sample times for %d timeslots"%len(times), file=log)
        dt = self.time_inc*60
        beam_times = [ times[0] ]
        for t in times[1:]:
            if t - beam_times[-1] >= dt:
                beam_times.append(t)
        if not quiet:
            print("  DtBeamMin=%.2f min results in %d samples"%(self.time_inc, len(beam_times)), file=log)
        if self.pa_inc:
            pas = [ 
                # put antenna0 position as reference frame. NB: in the future may want to do it per antenna
                dm.do_frame(self.pos0) and 
                # put time into reference frame
                dm.do_frame(dm.epoch("UTC",dq.quantity(t0,"s"))) and
                # compute PA 
                dm.posangle(self.field_centre,self.zenith).get_value("deg") for t0 in beam_times ]
            pa0 = pas[0]
            beam_times1 = [ beam_times[0] ]
            for t, pa in zip(beam_times[1:], pas[1:]):
                if abs(pa-pa0) >= self.pa_inc:
                    beam_times1.append(t)
                    pa0 = pa
            if not quiet:
                print("  FITSParAngleIncrement=%.2f deg results in %d samples"%(self.pa_inc, len(beam_times1)), file=log)
            beam_times = beam_times1
        beam_times.append(times[-1]+1)
        return beam_times

    def getFreqDomains (self):
        domains = np.zeros((len(self.freqs),2),np.float64)
        df = (self.freqs[1]-self.freqs[0])/2 if len(self.freqs)>1 else self.freqs[0]
        domains[:,0] = self.freqs-df
        domains[:,1] = self.freqs+df
#        import pdb; pdb.set_trace()
        return domains

    def radec2lm_scalar(self,ra,dec):
        """ 
            SIN projected world to l,m projection 
            AIPS memo series, 27
            Eric Greisen
        """
        ra0, dec0 = self.field_centre_radec
        l = np.cos(dec) * np.sin(ra - ra0)
        m = np.sin(dec) * np.cos(dec0) - np.cos(dec) * np.sin(dec0) * np.cos(ra - ra0)
        return l,m

    def evaluateBeam (self, t0, ra, dec):
        """Evaluates beam at time t0, in directions ra, dec.
        Inputs: t0 is a single time. ra, dec are Ndir vectors of directions.
        Output: a complex array of shape [Ndir,Nant,Nfreq,2,2] giving the Jones matrix per antenna, direction and frequency
        """

        # put antenna0 position as reference frame. NB: in the future may want to do it per antenna
        dm.do_frame(self.pos0);
        # put time into reference frame
        dm.do_frame(dm.epoch("UTC",dq.quantity(t0,"s")))
        # compute PA 
        parad = dm.posangle(self.field_centre,self.zenith).get_value("rad")
        import math
        # print("time %f, position angle %f"%(t0, parad*180/math.pi), file=log)

        # compute l,m per direction
        ndir = len(ra)
        l0 = numpy.zeros(ndir,float)
        m0 = numpy.zeros(ndir,float)
        for i,(r1,d1) in enumerate(zip(ra,dec)):
            l0[i], m0[i] = self.radec2lm_scalar(r1,d1)
        # print(ra*180/np.pi,dec*180/np.pi, file=log)
        # print(l0*180/np.pi,m0*180/np.pi, file=log)
        # rotate each by parallactic angle
        r = numpy.sqrt(l0*l0+m0*m0)
        angle = numpy.arctan2(m0,l0)
        l = r*numpy.cos(angle + parad + np.deg2rad(self.feedangle))
        m = r*numpy.sin(angle + parad + np.deg2rad(self.feedangle))



        # # print("Beam evaluated for l,m", file=log)
        # print(l, file=log)
        # print(m, file=log)
        # print("time %f, position angle %f"%(t0, parad*180/math.pi), file=log)

        # get interpolated values. Output shape will be [ndir,nfreq]
        if self.use_unity_ejones:
            beamjones = [ np.ones([ndir, len(self.freqs)], dtype=numpy.complex64),
                          np.zeros([ndir, len(self.freqs)], dtype=numpy.complex64),
                          np.zeros([ndir, len(self.freqs)], dtype=numpy.complex64),
                          np.ones([ndir, len(self.freqs)], dtype=numpy.complex64) ]
        else:
            beamjones = [ self.vbs[corr].interpolate(l,m,freq=self.freqs,freqaxis=1) for corr in self.corrs ]

        # now make output matrix

        jones = numpy.zeros((ndir,self.station_positions.shape[0],len(self.freqs),2,2),dtype=numpy.complex64)
        # populate it with values
        # NB: here we copy the same Jones to every antenna. In principle we could compute
        # a parangle per antenna. When we have pointing error, it's also going to be per
        # antenna
        for iant in range(self.station_positions.shape[0]):
            for ijones,(ix,iy) in enumerate(((0,0),(0,1),(1,0),(1,1))):
                bj = beamjones[ijones]
                jones[:,iant,:,ix,iy] = beamjones[ijones].reshape((len(bj),1)) if bj.ndim == 1 else bj

        feedswap_jones = np.array([[1., 0.], [0., 1.]], dtype=numpy.complex64)
        if self.applyantidiagonal:
            feedswap_jones = np.array([[0., 1.], [1., 0.]], dtype=numpy.complex64)

        Pjones = np.array([[1., 0.], [0., 1.]], dtype=numpy.complex64)
        if self.applyrotation:
            print("Applying derotation to data, since beam is sampled in time. "
                  "If you have equatorial mounts this is not what you should be doing!", file=log)
            if self.feedbasis == "linear":
                """ 2D rotation matrix according to Hales, 2017: 
                Calibration Errors in Interferometric Radio Polarimetry """
                c1, s1 = np.cos(parad + np.deg2rad(self.feedangle)), np.sin(parad + np.deg2rad(self.feedangle))
                # assume all stations has same parallactic angle
                Pjones[0, 0] = c1
                Pjones[0, 1] = s1
                Pjones[1, 0] = -s1
                Pjones[1, 1] = c1
            elif self.feedbasis == "circular":
                """ phase rotation matrix according to Hales, 2017: 
                Calibration Errors in Interferometric Radio Polarimetry """
                e1 = np.exp(1.0j * -(parad + np.deg2rad(self.feedangle)))
                e2 = np.exp(1.0j * (parad + np.deg2rad(self.feedangle)))
                # assume all stations has same parallactic angle
                Pjones[0, 0] = e1
                Pjones[0, 1] = 0
                Pjones[1, 0] = 0
                Pjones[1, 1] = e2
            else:
                raise RuntimeError("Feed basis not supported")

        # dot diagonal block matrix of P with E diagonal block vector 
        # again assuming constant P matrix across all stations
        E_vec = jones.reshape(ndir * self.station_positions.shape[0] * len(self.freqs), 2, 2)
        for i in range(E_vec.shape[0]):
            E_vec[i,:,:] = np.dot(Pjones, np.dot(E_vec[i, :, :], feedswap_jones))

        return E_vec.reshape(ndir, self.station_positions.shape[0], len(self.freqs), 2, 2)










import numpy as np
import scipy.stats as sstats
import scipy.signal as ssig
import scipy.spatial as spat
import copy

import time
def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print(('%r  %2.2f ms' % \
                  (method.__name__, (te - ts) * 1000)))
        
        return result
    return timed

DEBUG = True

class BoundingConvexHull(object):
    def __init__(self, list_hulls, name="unnamed", mask = None, check_mask_outofbounds=True):
        """ Initializes a bounding convex hull around a list of bounding convex hulls or series of
            points. A unity-weighted mask is computed for the region that falls within this convex hull
            if a mask of (y, x) coordinates is not provided. Otherwise if a mask is provided and the 
            check_mask_outofbounds value is set the masked coordinates are not verified to fall within
            the hull. The latter should thus be used with some caution by the user, but can potentially
            significantly speed up the mask creation process for axis aligned regions.
        """
        self._name = name
        self._check_mask_outofbounds = check_mask_outofbounds
        self._cached_filled_mask = None
        self._vertices = points = np.vstack([b.corners
            if hasattr(b, "corners") else [b[0], b[1]] for b in list_hulls])
        self._hull = spat.ConvexHull(points)
        if mask is None:
            self._mask, self._mask_weights = self.init_mask()
        else: 
            self.sparse_mask = mask

    def invalidate_cached_masks(self):
        """ Invalidates the cached masks (sparse or regular) """
        self._cached_filled_mask = None
        self._mask, self._mask_weights = self.init_mask()

    def __str__(self):
        return ",".join(["({0:d},{1:d})".format(x,y) for (x,y) in self.corners])
    
    def init_mask(self):
        """ creates a sparse mask of the convex hull of the form (y, x) tuples """
        lines = np.hstack([self.corners, np.roll(self.corners, -1, axis=0)])
        minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
        miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
        x = np.arange(minx, maxx + 1, 1) #upper limit inclusive
        y = np.arange(miny, maxy + 1, 1)
        meshgrid = np.meshgrid(y, x)
        bounding_mesh = list(zip(*[np.ravel(x) for x in np.meshgrid(y, x)]))
        
        sparse_mask = bounding_mesh if not self._check_mask_outofbounds else \
            [c for c in bounding_mesh if c[::-1] in self]
        
        mask_weights = np.ones(len(sparse_mask)) #initialize to unity, this should be modified when coadding
        return sparse_mask, mask_weights

    @property
    def sprase_mask_weights(self):
        """ returns sparse mask weights """
        return self._mask_weights

    @property
    def sparse_mask(self):
        """ returns a sparse mask (y, x) values of all points in the masked region """
        return self._mask

    @sparse_mask.setter
    def sparse_mask(self, mask):
        """ Sets the mask of the hull from a sparse mask - list of (y, x) coordinates """
        if not isinstance(mask, list):
            raise TypeError("Mask must be list")
        if not (hasattr(mask, "__len__") and (len(mask) == 0 or (hasattr(mask[0], "__len__") and len(mask[0]) == 2))):
            raise TypeError("Mask must be a sparse mask of 2 element values")
        if self._check_mask_outofbounds:
            self._mask = copy.deepcopy([c for c in mask if (c[1], c[0]) in self])
        else:
            self._mask = copy.deepcopy(mask)
        self._mask_weights = np.ones(len(self._mask))
    
    @property
    def mask(self, dtype=np.float64):
        """ Creates a filled rectangular mask grid of size y, x """
        if self._cached_filled_mask is not None:
            return self._cached_filled_mask

        lines = np.hstack([self.corners, np.roll(self.corners, -1, axis=0)])
        minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
        miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
        nx = maxx - minx + 1 # inclusive
        ny = maxy - miny + 1
        mesh = np.zeros(nx*ny, dtype=dtype)
        if nx==0 or ny==0 or len(self.sparse_mask) == 0: 
            self._cached_filled_mask = mesh.reshape((ny, nx))
        else:
            sparse_mask = np.array(self.sparse_mask)
            sel = np.logical_and(np.logical_and(sparse_mask[:, 1] >= minx,
                                                sparse_mask[:, 1] <= maxx),
                                 np.logical_and(sparse_mask[:, 0] >= miny,
                                                sparse_mask[:, 0] <= maxy))
            flat_index = (sparse_mask[sel][:, 0] - miny)*nx + (sparse_mask[sel][:, 1] - minx)
            mesh[flat_index] = self._mask_weights[sel]
            self._cached_filled_mask = mesh.reshape((ny, nx))
        return self._cached_filled_mask

    @classmethod
    def regional_data(cls, sel_region, data_cube, axes=(2, 3), oob_value=0):
        """ 2D array containing all values within convex hull 
            sliced out along axes provided as argument. Portions of sel_region
            that are outside of the data_cube is set to oob_value

            assumes the last value of axes is the fastest varying axis
        """
        if not isinstance(sel_region, BoundingConvexHull):
            raise TypeError("Object passed in is not of type BoundingConvexHull")
        if not (hasattr(axes, "__len__") and len(axes) == 2):
            raise ValueError("Expected a tupple of axes along which to slice out a region")
        axes = sorted(axes)

        lines = np.hstack([sel_region.corners, np.roll(sel_region.corners, -1, axis=0)])
        minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
        miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
        x = np.arange(minx, maxx + 1, 1)
        y = np.arange(miny, maxy + 1, 1)

        pad_left = max(0, 0 - minx)
        pad_bottom = max(0, 0 - miny)
        pad_right = max(0, maxx - data_cube.shape[axes[1]] + 1) #inclusive of upper limit
        pad_top = max(0, maxy - data_cube.shape[axes[0]] + 1)

        if minx > data_cube.shape[axes[0]] or miny > data_cube.shape[axes[1]] or \
           maxy < 0 or maxx < 0:
           raise ValueError("Expected a bounding hull that is at least partially within the image")

        # extract data, pad if necessary
        slc_data = [slice(None)] * len(data_cube.shape)
        for (start, end), axis in zip([(miny + pad_bottom, maxy - pad_top + 1), 
                                       (minx + pad_left, maxx - pad_right + 1)], axes):
            slc_data[axis] = slice(start, end)
        slc_padded = [slice(None)] * len(data_cube.shape)
        for (start, end), axis in zip([(pad_bottom, -miny + maxy + 1 - pad_top), 
                                       (pad_left, -minx + maxx + 1 - pad_right)], axes):
            slc_padded[axis] = slice(start, end)

        selected_data = data_cube[tuple(slc_data)]
        new_shape = list(data_cube.shape)
        new_shape[axes[0]] = (maxy - miny + 1)
        new_shape[axes[1]] = (maxx - minx + 1)

        if any(np.array([pad_left, pad_bottom, pad_right, pad_top]) > 0):
            padded_data = np.zeros(tuple(new_shape), dtype=selected_data.dtype) * oob_value
            padded_data[tuple(slc_padded)] = selected_data.copy()
        else:
            padded_data = selected_data.copy()

        # finally apply mask
        slc_padded_data = [slice(None)] * len(padded_data.shape)
        for (start, end), axis in zip([(0, maxy - miny + 1),  #mask starts at origin in the padded image
                                       (0, maxx - minx + 1)], axes):
            slc_padded_data[axis] = slice(start, end)
        slc_mask = [None] * len(padded_data.shape)
        for (start, end), axis in zip([(0, sel_region.mask.shape[0]),  #mask starts at origin in the padded image
                                       (0, sel_region.mask.shape[1])], axes):
            slc_mask[axis] = slice(start, end)
        mask = sel_region.mask.copy()
        mask[mask == 0] = oob_value
        padded_data[tuple(slc_padded_data)] *= mask[tuple(slc_mask)]
        window_extents = [minx, maxx, 
                          miny, maxy]
        return padded_data, window_extents
    
    @classmethod
    def normalize_masks(cls, regions, only_overlapped_regions=True):
        """ Normalizes region masks for overlapping pixels. This is necessary to properly coadd
            overlapping facets. If masks are guarenteed to be initialized to unity (e.g. after
            bounding region creation) the user can skip normalizing non-overlapping regions with
            flag only_overlapped_regions.
        """
        if not all([isinstance(reg, BoundingConvexHull) for reg in regions]):
            raise TypeError("Expected a list of bounding convex hulls")
        # Implements painters-like algorithm to
        # count the number of times a pixel coordinate falls within masks
        # The overlapping sections of regions can then be normalized
        # For now all regions have equal contribution
        allmasks = []
        for reg in regions:
            allmasks += list(reg.sparse_mask) if isinstance(reg.sparse_mask, np.ndarray) else reg.sparse_mask
        
        # flatten for faster comparisons
        allmasks = np.array(allmasks)
        maxx = np.max(allmasks[:, 1])
        nx = maxx + 1
        allmasks_flatten = allmasks[:, 0] * nx + allmasks[:, 1]

        # now count the number of times a pixel is painted onto
        unique_pxls_flatten, paint_count = np.unique(allmasks_flatten, return_counts=True)
        paint_count = paint_count.astype(np.float)
        
        if only_overlapped_regions:
            sel = paint_count > 1
            unique_pxls_flatten = unique_pxls_flatten[sel]
            paint_count = paint_count[sel]
        
        # with the reduced number of overlap pixels unflatten
        unique_pxls = np.vstack([unique_pxls_flatten // nx,
                                 unique_pxls_flatten % nx]).T
        unique_pxls = list(map(tuple, unique_pxls))
        paint_count[...] = 1.0 / paint_count

        # and finally update mask weights
        for reg in regions:
            reg._cached_filled_mask = None # invalidate
            overlap = [x for x in zip(paint_count, unique_pxls) if x[1] in reg.sparse_mask]
            for px_pc, px in overlap:
                sel = reg.sparse_mask.index(px) if isinstance(reg.sparse_mask, list) else \
                      np.all(reg.sparse_mask - px == 0, axis=1)
                reg._mask_weights[sel] = px_pc

    @property
    def circumference(self):
        """ area contained in hull """
        lines = self.edges
        return np.sum(np.linalg.norm(lines[:, 1, :] - lines[:, 0, :], axis=1) + 1)
    @property
    def area(self):
        """ area contained in hull """
        lines = np.hstack([self.corners, np.roll(self.corners, -1, axis=0)])
        return 0.5 * np.abs(np.sum([x1*(y2)-(x2)*y1 for x1,y1,x2,y2 in lines])) + 0.5 * self.circumference - 1

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, v):
        self._name = v

    @property
    def corners(self):
        """ Returns vertices and guarentees clockwise winding """
        return self._vertices[self._hull.vertices][::-1]

    def normals(self, left = True):
        """ return a list of left normals to the hull """
        normals = []
        for i in range(self.corners.shape[0]):
            # assuming clockwise winding
            j = (i + 1) % self.corners.shape[0]
            edge = self.corners[j, :] - self.corners[i, :]
            if left:
                normals.append((-edge[1], edge[0]))
            else:
                normals.append((edge[1], -edge[0]))
        return np.asarray(normals, dtype=np.double)

    @property
    def edges(self):
        """ return edge segments of the hull (clockwise wound) """
        edges = []
        for i in range(self.corners.shape[0]):
            # assuming clockwise winding
            j = (i + 1) % self.corners.shape[0]
            edge = tuple([self.corners[i, :], self.corners[j, :]])
            edges.append(edge)
        return np.asarray(edges, dtype=np.double)

    @property
    def edge_midpoints(self):
        """ return edge midpoints of the hull (clockwise wound) """
        edges = self.edges
        return np.mean(edges, axis=1)

    @property
    def lnormals(self):
        """ left normals to the edges of the hull """
        return self.normals(left = True)

    @property
    def rnormals(self):
        """ right normals to the edges of the hull """
        return self.normals(left=False)
    
    def overlaps_with(self, other, min_sep_dist=0.5): #less than half a pixel away
        """ 
            Implements the separating lines collision detection theorem 
            to test whether the hull intersects with 'other' hull
        """
        if not isinstance(other, BoundingConvexHull):
            raise TypeError("rhs must be a BoundingConvexHull")

        # get the projection axes
        normals = np.vstack([self.lnormals, other.lnormals])
        norms = np.linalg.norm(normals, axis=1)
        normals = normals / norms[None, 2]

        # compute vectors to corners from origin
        vecs_reg1 = self.corners
        vecs_reg2 = other.corners

        # compute projections onto normals
        for ni, n in enumerate(normals):
            projs = np.dot(vecs_reg1, n.T)
            minproj_reg1 = np.min(projs)
            maxproj_reg1 = np.max(projs)
            projs = np.dot(vecs_reg2, n.T)
            minproj_reg2 = np.min(projs)
            maxproj_reg2 = np.max(projs)
            if minproj_reg2 - maxproj_reg1 > min_sep_dist or minproj_reg1 - maxproj_reg2 > min_sep_dist:
                return False
        return True

    @property
    def centre(self, integral=True):
        """ Barycentre of hull """
        if integral:
            def rnd(x):
                return int(np.floor(x) if x >= 0 else np.ceil(x))
            return [rnd(x) for x in np.mean(self._vertices, axis=0)]
        else:
            return np.mean(self._vertices, axis=0)

    def __contains__(self, s, tolerance=0.5): #less than half a pixel away
        """ tests whether a point s(x,y) is in the convex hull """
        # there are three cases to consider
        # CASE 1:
        # scalar projection  between all inner pointing right normals (clockwise winding) 
        # and the point must be positive if the point were to lie inside
        # the region (true)
        # CASE 2:
        # point is on an edge - the scalar projection onto the axis is 0 for that edge
        # and greater than 0 for the other edges (true)
        # CASE 3:
        # it is outside (false)
        x, y = s
        isin = True
        normals = self.rnormals 
        xyvec = np.array([x, y])[None, :] - np.array(self.corners)
        
        dot = np.einsum("ij,ij->i", normals, xyvec)
        return np.all(dot > -tolerance)

class BoundingBox(BoundingConvexHull):
    def __init__(self, xl, xu, yl, yu, name="unnamed", mask=None, **kwargs):
        if not all([isinstance(x, (int, np.int64, np.int32, np.int16)) for x in [xl, xu, yl, yu]]):
            raise ValueError("Box limits must be integers")
        self.__xnpx = abs(xu - xl + 1) #inclusive of the upper pixel
        self.__ynpx = abs(yu - yl + 1)
        BoundingConvexHull.__init__(self,
                                    [[xl,yl],[xl,yu],[xu,yu],[xu,yl]],
                                    name,
                                    mask=mask,
                                    **kwargs)

    def init_mask(self):
        """ creates a sparse mask of the convex hull of the form (y, x) tuples """
        lines = np.hstack([self.corners, np.roll(self.corners, -1, axis=0)])
        minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
        miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
        x = np.arange(minx, maxx + 1, 1) #upper limit inclusive
        y = np.arange(miny, maxy + 1, 1)
        meshgrid = np.meshgrid(y, x)
        bounding_mesh = list(zip(*[np.ravel(x) for x in np.meshgrid(y, x)]))
        
        sparse_mask = np.asarray(bounding_mesh) # by default for a BB region the mask is always going to be the entire region
        
        mask_weights = np.ones(len(sparse_mask)) #initialize to unity, this should be modified when coadding
        return sparse_mask, mask_weights

    def __contains__(self, s):
        """ tests whether a point s(x,y) is in the box"""
        lines = np.hstack([self.corners, np.roll(self.corners, -1, axis=0)])
        minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
        miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
        return s[0] >= minx and s[0] <= maxx and s[1] >= miny and s[1] <= maxy

    @property
    def box_npx(self):
        return (self.__xnpx, self.__ynpx)

    @property
    def sparse_mask(self):
        """ returns a sparse mask (y, x) values of all points in the masked region """
        return self._mask

    @sparse_mask.setter
    def sparse_mask(self, mask):
        """ Sets the mask of the hull from a sparse mask - list of (y, x) coordinates """
        if not isinstance(mask, list) and not isinstance(mask, np.ndarray):
            raise TypeError("Mask must be list")
        if not (hasattr(mask, "__len__") and (len(mask) == 0 or (hasattr(mask[0], "__len__") and len(mask[0]) == 2))):
            raise TypeError("Mask must be a sparse mask of 2 element values")
        if mask == []:
            self._mask = []
        else:
            lines = np.hstack([self.corners, np.roll(self.corners, -1, axis=0)])
            minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
            miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
            nx = maxx - minx + 1 # inclusive
            ny = maxy - miny + 1
            sparse_mask = np.asarray(mask)
            sel = np.logical_and(np.logical_and(sparse_mask[:, 1] >= minx,
                                                sparse_mask[:, 1] <= maxx),
                                 np.logical_and(sparse_mask[:, 0] >= miny,
                                                sparse_mask[:, 0] <= maxy))
            self._mask = sparse_mask[sel]
            self._mask_weights = np.ones(len(self._mask))

    @classmethod
    def project_regions(cls, regional_data_list, regions_list, axes=(2, 3), dtype=np.float64, **kwargs):
        """ Projects individial regions back onto a single contiguous cube """
        if not (hasattr(regional_data_list, "__len__") and hasattr(regions_list, "__len__") and \
            len(regions_list) == len(regional_data_list)):
            raise TypeError("Region data list and regions lists must be lists of equal length")
        if not all([isinstance(x, np.ndarray) for x in regional_data_list]):
            raise TypeError("Region data list must be a list of ndarrays")
        if not all([isinstance(x, BoundingBox) for x in regions_list]):
            raise TypeError("Region list must be a list of Axis Aligned Bounding Boxes")
        if regions_list == []:
            return np.empty((0))
        if not all([reg.ndim == regional_data_list[0].ndim for reg in regional_data_list]):
            raise ValueError("All data cubes must be of equal dimension")
        axes = tuple(sorted(axes))

        minx = np.min([np.min(f.corners[:, 0]) for f in regions_list])
        maxx = np.max([np.max(f.corners[:, 0]) for f in regions_list])
        miny = np.min([np.min(f.corners[:, 1]) for f in regions_list])
        maxy = np.max([np.max(f.corners[:, 1]) for f in regions_list])
        npxx = maxx - minx + 1
        npxy = maxy - miny + 1
        global_offsetx = -minx #-min(0, minx)
        global_offsety = -miny #-min(0, miny)

        projected_image_size = list(regional_data_list[0].shape)
        projected_image_size[axes[0]] = npxy
        projected_image_size[axes[1]] = npxx
        stitched_img = np.zeros(tuple(projected_image_size), dtype=dtype)

        combined_mask = []
        for f, freg in zip(regional_data_list, regions_list):
            f[np.isnan(f)] = 0
            xl = max(0, global_offsetx+np.min(freg.corners[:, 0]))
            xu = min(global_offsetx+np.max(freg.corners[:, 0]) + 1, npxx)
            yl = max(0, global_offsety+np.min(freg.corners[:, 1]))
            yu = min(global_offsety+np.max(freg.corners[:, 1]) + 1, npxy)
            fnx = xu - xl + 1 # inclusive
            fny = yu - yl + 1 # inclusive
            if f.shape[axes[0]] != fny - 1 or f.shape[axes[1]] != fnx - 1:
                raise ValueError("One or more bounding box descriptors does not match shape of corresponding data cubes")
            slc_data = [slice(None)] * len(stitched_img.shape)
            for (start, end), axis in zip([(yl, yu), (xl, xu)], axes):
                slc_data[axis] = slice(start, end)
                
            stitched_img[tuple(slc_data)] += f
            combined_mask += list(freg.sparse_mask)

        return stitched_img, BoundingBox(minx, maxx, miny, maxy, mask=combined_mask, **kwargs)

########################################################################
## Factories
########################################################################

class BoundingBoxFactory(object):
    @classmethod                                
    def AxisAlignedBoundingBox(cls, convex_hull_object, square=False, enforce_odd=True, **kwargs):
        """ Constructs an axis aligned bounding box around convex hull """
        if not isinstance(convex_hull_object, BoundingConvexHull):
            raise TypeError("Convex hull object passed in constructor is not of type BoundingConvexHull")
        if square:
            nx = np.max(convex_hull_object.corners[:, 0]) - np.min(convex_hull_object.corners[:, 0]) + 1 #inclusive
            ny = np.max(convex_hull_object.corners[:, 1]) - np.min(convex_hull_object.corners[:, 1]) + 1 #inclusive
            boxdiam = max(nx, ny)
            boxrad = boxdiam // 2
            cx, cy = convex_hull_object.centre
            xl = cx - boxrad
            xu = cx + boxdiam - boxrad - 1
            yl = cy - boxrad
            yu = cy + boxdiam - boxrad - 1
        else:
            xl = np.min(convex_hull_object.corners[:, 0])
            xu = np.max(convex_hull_object.corners[:, 0])
            yl = np.min(convex_hull_object.corners[:, 1])
            yu = np.max(convex_hull_object.corners[:, 1])

        xu += (xu - xl) % 2 if enforce_odd else 0
        yu += (yu - yl) % 2 if enforce_odd else 0

        return BoundingBox(xl, xu, yl, yu,
                           convex_hull_object.name,
                           mask=convex_hull_object.sparse_mask, 
                           **kwargs)

    @classmethod
    def SplitBox(cls, bounding_box_object, nsubboxes=1, **kwargs):
        """ Split a axis-aligned bounding box into smaller boxes """
        if not isinstance(bounding_box_object, BoundingBox):
            raise TypeError("Expected bounding box object")
        if not (isinstance(nsubboxes, int) and nsubboxes >= 1):
            raise ValueError("nsubboxes must be integral type and be 1 or more")
        xl = np.min(bounding_box_object.corners[:, 0])
        xu = np.max(bounding_box_object.corners[:, 0])
        yl = np.min(bounding_box_object.corners[:, 1])
        yu = np.max(bounding_box_object.corners[:, 1])
        
        # construct a nonregular meshgrid bound to xu and yu
        x = xl + np.arange(0, nsubboxes + 1) * int(np.ceil((xu - xl + 1) / float(nsubboxes)))
        y = yl + np.arange(0, nsubboxes + 1) * int(np.ceil((yu - yl + 1) / float(nsubboxes)))
        xx, yy = np.meshgrid(x, y)

        # split into boxes
        xls = xx[0:-1, 0:-1].copy()
        xus = xx[1:, 1:].copy()
        yls = yy[0:-1, 0:-1].copy()
        yus = yy[1:, 1:].copy()

        # make sure no boxes overlap
        xus = xus - 1
        yus = yus - 1

        # clamp the final coordinate to the upper end (may result in rectanglular box at the end)
        xus[:, -1] = max(xu, min(xus[0, -1], xu)) 
        yus[-1, :] = max(yu, min(yus[-1, 0], yu))
        
        #coordinates for all the contained boxes, anti-clockwise wound
        xls = xls.ravel()
        yls = yls.ravel()
        xus = xus.ravel()
        yus = yus.ravel()
        bl = list(zip(xls, yls))
        br = list(zip(xus, yls))
        ur = list(zip(xus, yus))
        ul = list(zip(xls, yus))
        contained_boxes = list(zip(bl, br, ur, ul))

        #finally create bbs for each of the contained boxes with the mask
        #chopped up between the boxes by the convex hull initializer
        new_regions = [BoundingBox(bl[0], br[0], bl[1], ul[1],
                                   bounding_box_object.name,
                                   mask=bounding_box_object.sparse_mask, 
                                   **kwargs)
                       for bl, br, ur, ul in contained_boxes]

        return new_regions

    @classmethod
    def PadBox(cls, bounding_box_object, desired_nx, desired_ny, **kwargs):
        """ Creates a box with a padded border around a axis-aligned bounding box """
        if not isinstance(bounding_box_object, BoundingBox):
            raise TypeError("Expected bounding box object")
        nx, ny = bounding_box_object.box_npx
        if desired_nx - nx < 0 or desired_ny - ny < 0:
            raise ValueError("Padded box must be bigger than original box")
        pad_left = desired_nx // 2
        pad_right = desired_nx - pad_left - 1
        pad_bottom = desired_ny // 2
        pad_top = desired_ny - pad_bottom - 1
        cx, cy = bounding_box_object.centre
        xl = cx - pad_left
        xu = cx + pad_right
        yl = cy - pad_bottom
        yu = cy + pad_top
        return BoundingBox(xl, xu, yl, yu,
                           bounding_box_object.name, 
                           mask=bounding_box_object.sparse_mask, 
                           **kwargs) #mask unchanged in the new shape, border frame discarded




from __future__ import print_function
from builtins import range
import numpy as np
import os, os.path
import traceback
from cubical.tools import logger, ModColor
from cubical.flagging import FL
from cubical.statistics import SolverStats
from cubical.tools import BREAK  # useful: can set static breakpoints by putting BREAK() in the code
from cubical.madmax import plots

log = logger.getLogger("madmax")


# Conversion factor for sigma = SIGMA_MAD*mad
SIGMA_MAD = 1.4826

import builtins
try:
    builtins.profile
except AttributeError:
    # No line profiler, provide a pass-through version
    def profile(func): return func
    builtins.profile = profile 
    

class Flagger(object):
    def __init__(self, GD, chunk_label, metadata, stats):
        self.GD = GD
        self.metadata = metadata
        self.stats = stats
        self.chunk_label = chunk_label

        self.flag_warning_threshold = GD['flags']["warn-thr"]

        self.mad_threshold = GD['madmax']['threshold']
        self.medmad_threshold = GD['madmax']['global-threshold']
        
        # make copies of theses lists, since we manipuate them in get_thresholds()
        if not isinstance(self.mad_threshold, list):
            self.mad_threshold = [self.mad_threshold]
        else:
            self.mad_threshold = list(self.mad_threshold)
        if not isinstance(self.medmad_threshold, list):
            self.medmad_threshold = [self.medmad_threshold]
        else:
            self.medmad_threshold = list(self.medmad_threshold)
            
        self.mad_diag = GD['madmax']['diag']
        self.mad_offdiag = self.metadata.num_corrs == 4 and GD['madmax']['offdiag']
        if not self.mad_diag and not self.mad_offdiag:
            self._mode = False

        # setup MAD estimation settings
        self.mad_per_corr = False
        if GD['madmax']['estimate'] == 'corr':
            self.mad_per_corr = True
            self.mad_estimate_diag, self.mad_estimate_offdiag = self.mad_diag, self.mad_offdiag
        elif GD['madmax']['estimate'] == 'all':
            self.mad_estimate_diag = True
            self.mad_estimate_offdiag = self.metadata.num_corrs == 4
        elif GD['madmax']['estimate'] == 'diag':
            self.mad_estimate_diag, self.mad_estimate_offdiag = True, False
        elif GD['madmax']['estimate'] == 'offdiag':
            if self.metadata.num_corrs == 4:
                self.mad_estimate_diag, self.mad_estimate_offdiag = False, True
            else:
                self.mad_estimate_diag, self.mad_estimate_offdiag = True, False
        else:
            raise RuntimeError("invalid --madmax-estimate {} setting".format(GD['madmax']['estimate']))

        if GD['madmax']['plot-bl']:
            self._plot_baselines = set(GD['madmax']['plot-bl'].split(","))
        else:
            self._plot_baselines = set()

        self._plotnum = 0

    def set_mode(self, mode):
        """Sets operational mode for flagger"""
        self._mode = mode
        self._pretend = self._mode == "pretend"
        self.trial_mode = self._trial = self._mode == "trial"
        if self._pretend:
            self.desc_mode = "Pretend-Mad Max"
        elif self._mode == "trial":
            self.desc_mode = "Trial-Mad Max"
        elif self._mode:
            self.desc_mode = "Mad Max"
        else:
            self.desc_mode = "No Max"
        self.flagbit = 0 if self._pretend else FL.MAD

    def get_mad_thresholds(self):
        """MAD thresholds above are either a list, or empty. Each time we access the list, we pop the first element,
        until the list is down to one element."""
        if not self._mode:
            return 0, 0
        return self.mad_threshold.pop(0) if len(self.mad_threshold) > 1 else \
                   (self.mad_threshold[0] if self.mad_threshold else 0), \
               self.medmad_threshold.pop(0) if len(self.medmad_threshold) > 1 else \
                   (self.medmad_threshold[0] if self.medmad_threshold else 0)

    def get_plot_filename(self, kind=''):
        plotdir = '{}-madmax.plots'.format(self.GD['out']['name'])
        if not os.path.exists(plotdir):
            try:
                os.mkdir(plotdir)
            # allow a failure -- perhaps two workers got unlucky and both are trying to make the
            # same directory. Let savefig() below fail instead
            except OSError:
                pass
        if kind:
            filename = '{}/{}.{}.{}.png'.format(plotdir, self.chunk_label, self._plotnum, kind)
        else:
            filename = '{}/{}.{}.png'.format(plotdir, self.chunk_label, self._plotnum)
        self._plotnum += 1
        return filename

    @builtins.profile
    def report_carnage(self, absres, mad, baddies, flags_arr, method, max_label):
        made_plots = False
        n_tim, n_fre, n_ant, n_ant = baddies.shape
        nbad = int(baddies.sum())
        self.stats.chunk.num_mad_flagged += nbad

        if nbad or self._plot_baselines:
            if nbad < flags_arr.size * self.flag_warning_threshold:
                warning, color = "", "blue"
            else:
                warning, color = "WARNING: ", "red"
            frac = nbad / float(baddies.size)
            mode = "trial-" if self._trial else ("pretend-" if self._pretend else "")
            print("{warning}{max_label} {method} {mode}flags {nbad} ({frac:.2%}) visibilities".format(**locals()), file=log(1, color))
            if log.verbosity() > 2 or self.GD['madmax']['plot']:
                per_bl = []
                total_elements = float(n_tim * n_fre)
                interesting_fraction = self.GD['madmax']['plot-frac-above']*total_elements
                plot_explicit_baselines = []
                for p in range(n_ant):
                    for q in range(p + 1, n_ant):
                        n_flagged = baddies[:, :, p, q].sum()
                        if n_flagged and n_flagged >= interesting_fraction:
                            per_bl.append((n_flagged, p, q))
                        if self.metadata.baseline_name[p,q] in self._plot_baselines:
                            plot_explicit_baselines.append(((n_flagged, p, q), "--madmax-plot-bl"))
                per_bl = sorted(per_bl, reverse=True)
                # print
                per_bl_str = ["{} ({}m): {} ({:.2%})".format(self.metadata.baseline_name[p,q],
                                int(self.metadata.baseline_length[p,q]), n_flagged, n_flagged/total_elements)
                              for n_flagged, p, q in per_bl]
                print("{} of which per baseline: {}".format(max_label, ", ".join(per_bl_str)), file=log(3))
                # plot, if asked to
                if self.GD['madmax']['plot']:
                    baselines_to_plot = []
                    if len(per_bl):
                        baselines_to_plot.append((per_bl[0], "worst baseline"))
                    if len(per_bl)>2:
                        baselines_to_plot.append((per_bl[len(per_bl)//2], "median baseline"))
                    baselines_to_plot += plot_explicit_baselines
                    import pylab
                    for (n_flagged, p, q), baseline_label in baselines_to_plot:
                        blname = self.metadata.baseline_name[p, q] 
                        # make subplots
                        subplot_titles = {}
                        for c1,x1 in enumerate(self.metadata.feeds.upper()):
                            for c2,x2 in enumerate(self.metadata.feeds.upper()):
                                mm = mad[0,p,q,c1,c2] if self.mad_per_corr else mad[0,p,q]
                                subplot_titles[c1,c2] = "{}{} residuals (MAD {:.3g})".format(x1, x2, 0 if mm is np.ma.masked else mm) 
                        try:
                            figure = plots.make_dual_absres_plot(absres, flags_arr!=0, baddies, p, q, self.metadata, subplot_titles)
                            # make plot title with some info
                            fraction = n_flagged / total_elements
                            blname = self.metadata.baseline_name[p,q]
                            bllen  = int(self.metadata.baseline_length[p,q])
                            pylab.suptitle("{} {}: baseline {} ({}m), {} ({:.2%}) visibilities killed ({})".format(max_label,
                                            method, blname, bllen, n_flagged, fraction, baseline_label))
                            # save or show plot
                            if self.GD['madmax']['plot'] == 'show':
                                pylab.show()
                            else:
                                filename = self.get_plot_filename()
                                figure.savefig(filename, dpi=300)
                                print("{}: saving Mad Max flagging plot to {}".format(self.chunk_label,filename), file=log(1))
                            pylab.close(figure)
                            del figure
                            made_plots = True
                        except Exception as exc:
                            if self.GD['debug']['escalate-warnings']:
                                raise
                            traceback.print_exc()
                            print("WARNING: {}: exception {} raised while generating Mad Max waterfall plot for baseline {} ({})".format(
                                            self.chunk_label, exc, blname, baseline_label), file=log(0, "red"))
                            print("Although harmless, this may indicate a problem with the data, or a bug in CubiCal.", file=log(0))
                            print("Please see stack trace above, and report if you think this is a bug.", file=log(0))
        else:
            print("{} {} abides".format(max_label, method), file=log(2))

        return made_plots, nbad>0


    @builtins.profile
    def beyond_thunderdome(self, resid_arr, data_arr, model_arr, flags_arr, threshold, med_threshold, max_label):
        """This function implements MAD-based flagging on residuals"""
        if not threshold and not med_threshold:
            return False
        # residuals can pe per-model (when invoked from solver) or not, when invoked on final residuals.
        # cope with both cases
        if resid_arr.ndim == 7:
            n_mod, _, _, n_ant, n_ant, n_cor, n_cor = resid_arr.shape
        else:
            _, _, n_ant, n_ant, n_cor, n_cor = resid_arr.shape
            n_mod = 1
            resid_arr = resid_arr.reshape([1]+list(resid_arr.shape))

        import cubical.kernels
        madmax = cubical.kernels.import_kernel("madmax")
        # estimate MAD of off-diagonal elements
        absres = np.empty_like(resid_arr, dtype=np.float32)
        np.abs(resid_arr, out=absres)
        if self.mad_per_corr:
            mad, goodies = madmax.compute_mad_per_corr(absres, flags_arr, diag=self.mad_estimate_diag, offdiag=self.mad_estimate_offdiag)
        else:
            mad, goodies = madmax.compute_mad(absres, flags_arr, diag=self.mad_estimate_diag, offdiag=self.mad_estimate_offdiag)
        # any of it non-zero?
        if mad.mask.all():
            return
        # estimate median MAD
        ### want to do this:
        ## medmad = np.ma.median(mad, axis=(1,2))
        ### but this seems to thrown an error on earlier numpys (1.11?), so let's be defensive and reshape into one axis:
        shape1 = [mad.shape[0], mad.shape[1]*mad.shape[2]] + list(mad.shape[3:])
        medmad = np.ma.median(mad.reshape(shape1), axis=1)
        # all this was worth it, just so I could type "mad.max()" as legit code
        print("{} per-baseline MAD min {:.3g}, max {:.3g}, median {:.3g} to {:.3g}".format(max_label, mad.min(), mad.max(), medmad.min(), medmad.max()), file=log(2))
        if log.verbosity() > 4:
            for imod in range(n_mod):
                if self.mad_per_corr:
                    for ic1,c1 in enumerate(self.metadata.feeds):
                        for ic2,c2 in enumerate(self.metadata.feeds):
                            per_bl = [(mad[imod,p,q,ic1,ic2], p, q) for p in range(n_ant)
                                      for q in range(p+1, n_ant) if not mad.mask[imod,p,q,ic1,ic2]]
                            per_bl = ["{} ({}m): {:.3g}".format(self.metadata.baseline_name[p,q], int(self.metadata.baseline_length[p,q]), x)
                                      for x, p, q in sorted(per_bl)[::-1]]
                            print("{} model {} {}{} MADs are {}".format(max_label, imod,
                                                                                c1.upper(), c2.upper(), ", ".join(per_bl)), file=log(4))
                else:
                    per_bl = [(mad[imod,p,q,], p, q) for p in range(n_ant)
                              for q in range(p+1, n_ant) if not mad.mask[imod,p,q]]
                    per_bl = ["{} ({}m) {:.3g}".format(self.metadata.baseline_name[p,q], int(self.metadata.baseline_length[p,q]), x)
                              for x, p, q in sorted(per_bl)[::-1]]
                    print("{} model {} MADs are {}".format(max_label, imod, ", ".join(per_bl)), file=log(4))


        made_plots = flagged_something = False

        thr = np.zeros((n_mod, n_ant, n_ant, n_cor, n_cor), dtype=np.float32)
        # apply per-baseline MAD threshold
        if threshold:
            if self.mad_per_corr:
                thr[:] = threshold * mad / SIGMA_MAD
            else:
                thr[:] = threshold * mad[...,np.newaxis,np.newaxis] / SIGMA_MAD
            baddies = madmax.threshold_mad(absres, thr, flags_arr, self.flagbit, goodies,
                                             diag=self.mad_diag, offdiag=self.mad_offdiag)
            made_plots, flagged_something  = self.report_carnage(absres, mad, baddies, flags_arr,
                                                "baseline-based Mad Max ({} sigma)".format(threshold), max_label)
            if not self._pretend:
                baddies = baddies.astype(bool)
                if model_arr is not None:
                    model_arr[:,:,baddies,:,:] = 0
                if data_arr is not None:
                    data_arr[:,baddies,:,:] = 0

        # apply global median MAD threshold
        if med_threshold:
            med_thr = med_threshold * medmad / SIGMA_MAD
            if self.mad_per_corr:
                thr[:] = med_thr[:,np.newaxis,np.newaxis,:,:]
            else:
                thr[:] = med_thr[:,np.newaxis,np.newaxis,np.newaxis,np.newaxis]
            baddies = madmax.threshold_mad(absres, thr, flags_arr, self.flagbit, goodies,
                                             diag=self.mad_diag, offdiag=self.mad_offdiag)

            made, flagged = \
                self.report_carnage(absres, mad, baddies, flags_arr,
                                       "global Mad Max ({} sigma)".format(med_threshold), max_label)

            made_plots = made_plots or made
            flagged_something = flagged_something or flagged

            if not self._pretend:
                baddies = baddies.astype(bool)
                if model_arr is not None:
                    model_arr[:, :, baddies, :, :] = 0
                if data_arr is not None:
                    data_arr[:, baddies, :, :] = 0
        else:
            med_thr = None

        # generate overview plot
        if made_plots:
            import pylab
            outflags, figure = plots.make_baseline_mad_plot(mad, medmad, med_thr, metadata=self.metadata,
                                max_label=max_label, chunk_label=self.chunk_label,
                                antenna_mad_threshold=self.GD['madmax']['flag-ant-thr'])
            if outflags.any():
                flagged_something = True
                if self.mad_per_corr:
                    outflags = outflags.any(axis=(-1,-2))
                if self.GD['madmax']['flag-ant'] and not self._pretend:
                    print("{} baselines {}flagged on mad residuals (--madmax-flag-ant 1)".format(
                                            outflags.sum()/2, "trial-" if self._trial else ""), file=log(0, "red"))
                    flags_arr[:,:,outflags] |= self.flagbit
                    if model_arr is not None:
                        model_arr[:,:,:,:,outflags,:,:] = 0
                    if data_arr is not None:
                        data_arr[:,:,:,outflags,:,:] = 0
                else:
                    print("{} baselines would have been flagged due to mad residuals (use --madmax-flag-ant to enable this)".format(outflags.sum()/2), file=log(0, "red"))

            try:
                if self.GD['madmax']['plot'] == 'show':
                    pylab.show()
                else:
                    filename = self.get_plot_filename('mads')
                    print("{}: saving MAD distribution plot to {}".format(self.chunk_label,filename), file=log(1))
                    figure.savefig(filename, dpi=300)
                    from future.moves import pickle
                    pickle_file = filename+".cp"
                    with open(pickle_file, "wb") as pf:
                        pickle.dump((mad, medmad, med_thr, self.metadata, max_label), pf, 2)
                    print("{}: pickled MAD distribution to {}".format(self.chunk_label, pickle_file), file=log(1))
                pylab.close(figure)
                del figure
            except Exception as exc:
                if self.GD['debug']['escalate-warnings']:
                    raise
                traceback.print_exc()
                print("WARNING: {}: exception {} raised while rendering Mad Max summary plot".format(
                                        self.chunk_label, exc), file=log(0,"red"))
                print("Although harmless, this may indicate a problem with the data, or a bug in CubiCal.", file=log(0))
                print("Please see stack trace above, and report if you think this is a bug.", file=log(0))

        return flagged_something and not self._pretend


# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from cubical.machines.interval_gain_machine import PerIntervalGains
import numpy as np
import numpy.ma as ma
from numpy.ma import masked_array
from cubical.flagging import FL
import cubical.kernels

from cubical.tools import logger
log = logger.getLogger("complex_2x2")


class Complex2x2Gains(PerIntervalGains):
    """
    This class implements the full complex 2x2 gain machine.
    """
    def __init__(self, label, data_arr, ndir, nmod, chunk_ts, chunk_fs, chunk_label, options):
        """
        Initialises a 2x2 complex gain machine.
        
        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            chunk_ts (np.ndarray):
                Times for the data being processed.
            chunk_fs (np.ndarray):
                Frequencies for the data being processsed.
            options (dict): 
                Dictionary of options. 
        """
        # note that this sets up self.kernel
        PerIntervalGains.__init__(self, label, data_arr, ndir, nmod,
                                  chunk_ts, chunk_fs, chunk_label, options)

        self._estimate_pzd = options["estimate-pzd"]

        # this will be set to PZD and exp(-i*PZD) once the PZD estimate is done
        self._pzd = 0
        self._exp_pzd = 1

    @classmethod
    def determine_diagonality(cls, options):
        """Returns true if the machine class, given the options, represents a diagonal gain"""
        updates = set(options['update-type'].split("-"))
        return options['type'] == 'complex-diag' or \
               ('full' not in updates and 'leakage' not in updates)

    @staticmethod
    def exportable_solutions():
        """ Returns a dictionary of exportable solutions for this machine type. """
        sols = PerIntervalGains.exportable_solutions()
        sols["pzd"] = (0.0, ("time", "freq"))
        return sols

    def importable_solutions(self, grid0):
        """ Returns a dictionary of importable solutions for this machine type. """
        sols = super(Complex2x2Gains, self).importable_solutions(grid0)
        if "pzd" in self.update_type:
            sols["pzd"] = dict(**self.interval_grid)
        return sols

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """
        sols = super(Complex2x2Gains, self).export_solutions()
        if "pzd" in self.update_type and self._pzd is not None:
            sols["pzd"] = (masked_array(self._pzd), self.interval_grid)
        return sols

    def import_solutions(self, soldict):
        """
        Loads solutions from a dict.

        Args:
            soldict (dict):
                Contains gains solutions which must be loaded.
        """
        if "pzd" in self.update_type and "pzd" in soldict:
            self._pzd = soldict["pzd"]
            self._exp_pzd = np.exp(-1j * self._pzd)

        super(Complex2x2Gains, self).import_solutions(soldict)


    def compute_js(self, obser_arr, model_arr):
        """
        This function computes the (J\ :sup:`H`\J)\ :sup:`-1` and J\ :sup:`H`\R terms of the GN/LM
        method.

        Args:
            obser_arr (np.ndarray):
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the
                observed visibilities.
            model_arr (np.ndrray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the
                model visibilities.

        Returns:
            3-element tuple

                - J\ :sup:`H`\R (np.ndarray)
                - (J\ :sup:`H`\J)\ :sup:`-1` (np.ndarray)
                - Count of flags raised (int)
        """

        n_dir, n_tim, n_fre, n_ant, n_cor, n_cor = self.gains.shape

        jh = self.get_new_jh(model_arr)

        self.kernel_solve.compute_jh(model_arr, self.gains, jh, self.t_int, self.f_int)

        jhr = self.get_new_jhr()
        r = self.get_obs_or_res(obser_arr, model_arr)

        if self.offdiag_only:
            jh[...,(0,1),(0,1)] = 0
            if r is obser_arr:
                r = r.copy()
            r[...,(0,1),(0,1)] = 0
            
        if self.diag_only:
            jh[...,(0,1),(1,0)] = 0
            if r is obser_arr:
                r = r.copy()
            r[...,(0,1),(1,0)] = 0

        self.kernel_solve.compute_jhr(jh, r, jhr, self.t_int, self.f_int)

        jhj, jhjinv = self.get_new_jhj()

        self.kernel_solve.compute_jhj(jh, jhj, self.t_int, self.f_int)

        if "scalar" in self.update_type:
            jhr[..., (0, 1), (0, 1)] = jhr[..., (0, 1), (0, 1)].mean(axis=-1)[..., np.newaxis]
            jhj[..., (0, 1), (0, 1)] = jhj[..., (0, 1), (0, 1)].mean(axis=-1)[..., np.newaxis]

        flag_count = self.kernel_solve.compute_jhjinv(jhj, jhjinv, self.gflags, self.eps, FL.ILLCOND)

#         if flag_count:
#             import pdb; pdb.set_trace()

        return jhr, jhjinv, flag_count

    #@profile
    def implement_update(self, jhr, jhjinv):

        # jhjinv is 2x2 block-diagonal, with Hermitian blocks. TODO: what's the variance on the off-diagonals?
        # variance of gain is diagonal of jhjinv
        if self.posterior_gain_error is None:
            self.posterior_gain_error = np.zeros_like(jhjinv.real)
        diag = jhjinv[..., (0, 1), (0, 1)].real
        self.posterior_gain_error[...,(0,1),(0,1)] = np.sqrt(diag)
        self.posterior_gain_error[...,(1,0),(0,1)] = np.sqrt(diag.sum(axis=-1)/2)[...,np.newaxis]

        update = self.init_update(jhr)

        self.kernel_solve.compute_update(jhr, jhjinv, update)

        if self.dd_term and self.n_dir > 1:
            update += self.gains

        self.restrict_solution(update)
        # raise flag so updates of G^H and G^-1 are computed
        self._gh_update = self._ghinv_update = True

        #if self._offdiag_only:
        #    update[...,0,0] = 1
        #    update[:,:,:,:,1,1] = self._exp_pzd[np.newaxis,:,:,np.newaxis]

        if self.iters % 2 == 0 or self.n_dir > 1:
            self.gains += update
            self.gains *= 0.5
            self.restrict_solution(self.gains)
        else:
            np.copyto(self.gains, update)

    def precompute_attributes(self, data_arr, model_arr, flags_arr, noise):
        """
        """
        super(Complex2x2Gains, self).precompute_attributes(data_arr, model_arr, flags_arr, noise)

        if self.solvable and self._estimate_pzd and self._pzd is 0 and model_arr is not None:
            marr = model_arr[..., (0, 1), (1, 0)][:, 0].sum(0)
            darr = data_arr[..., (0, 1), (1, 0)][0]
            mask = (flags_arr[..., np.newaxis] != 0) | (marr == 0)
            with np.errstate(divide='ignore', invalid='ignore'):
                dm = darr * (np.conj(marr) / abs(marr))
            dabs = np.abs(darr)
            dm[mask] = 0
            dabs[mask] = 0
            # collapse time/freq axis into intervals and sum antenna axes
            dm_sum = self.interval_sum(dm).sum(axis=(2, 3))
            dabs_sum = self.interval_sum(dabs).sum(axis=(2, 3))
            # sum off-diagonal terms
            dm_sum = dm_sum[..., 0] + np.conj(dm_sum[..., 1])
            dabs_sum = dabs_sum[..., 0] + np.conj(dabs_sum[..., 1])
            with np.errstate(divide='ignore',invalid='ignore'):
                pzd = np.angle(dm_sum / dabs_sum)
            pzd[dabs_sum == 0] = 0
            with np.printoptions(precision=4, suppress=True, linewidth=1000):
                print("{0}: PZD estimate {1} deg".format(self.chunk_label, pzd * 180 / np.pi), file=log(2))
            self._pzd = pzd
            self._exp_pzd = np.exp(-1j * pzd)

            self.gains[:, :, :, :, 0, 0] = 1
            self.gains[:, :, :, :, 1, 1] = self._exp_pzd[np.newaxis, :, :, np.newaxis]


    def restrict_solution(self, gains):
        """
        Restricts the solution by invoking the inherited restrict_soultion method and applying
        any machine specific restrictions.
        """

        if "pzd" in self.update_type:
            # re-estimate pzd
            mask = self.gflags!=0
            with np.errstate(divide='ignore', invalid='ignore'):
                pzd = masked_array(gains[:, :, :, :, 0, 0] / gains[:, :, :, :, 1, 1], mask)
            pzd = np.angle(pzd.sum(axis=(0,3)))
            with np.printoptions(precision=4, suppress=True, linewidth=1000):
                print("{0}: PZD estimate changes by {1} deg".format(self.chunk_label, (pzd-self._pzd)* 180 / np.pi), file=log(2))
            # import ipdb; ipdb.set_trace()
            self._pzd = pzd
            self._exp_pzd = np.exp(-1j * pzd)

            gains[:, :, :, :, 0, 0] = 1
            gains[:, :, :, :, 1, 1] = self._exp_pzd[np.newaxis, :, :, np.newaxis]
            
        if "leakage" in self.update_type:
            if "pzd" not in self.update_type:
                gains[:, :, :, :, 0, 0] = 1
                gains[:, :, :, :, 1, 1] = 1
                
            if "rel" in self.update_type and self.ref_ant is not None:
                offset =  gains[:, :, :, self.ref_ant, 0, 1].copy()
                gains[..., 0, 1] -= offset[..., np.newaxis]
                gains[..., 1, 0] += np.conj(offset)[..., np.newaxis]
                with np.printoptions(precision=4, suppress=True, linewidth=1000):
                    print("{0}: subtracting relative leakage offset {1}".format(self.chunk_label, offset), file=log(2))

        if self.ref_ant is not None:
            phase = np.angle(self.gains[...,self.ref_ant,0,0])
            gains[:,:,:,:,(0,1),(0,1)] *= np.exp(-1j*phase)[:,:,:,np.newaxis,np.newaxis]

        super(Complex2x2Gains, self).restrict_solution(gains)
        
        # with np.printoptions(precision=4, suppress=True):
        #     if self._jones_label == 'D':
        #         for p in range(self.n_ant):
        #             for a in (0,1):
        #                 for b in (0,1):
        #                     log.error("D{}{} for antenna {}: {}".format(a+1, b+1, p, gains[0,0,:,p,a,b]))

    @property
    def dof_per_antenna(self):
        if "leakage" in self.update_type and "pzd" in self.update_type:
            return 4 + 1./self.n_ant
        elif "leakage" in self.update_type:
            return 4
        elif "pzd" in self.update_type:
            return 1./self.n_ant
        else:
            return super(Complex2x2Gains, self).dof_per_antenna



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Main code body. Handles options, invokes solvers and manages multiprocessing.
"""

## OMS: this workaround should not be necessary, now that https://github.com/ratt-ru/CubiCal/issues/75 is fixed
# import logging
# if 'vext' in logging.Logger.manager.loggerDict.keys():
#     for handler in logging.root.handlers:
#         logging.root.removeHandler(handler)
#     logging.getLogger('vext').setLevel(logging.WARNING)
##
from __future__ import print_function
from builtins import range
from six import string_types
from future.moves import pickle
import os, os.path
import sys
import warnings
import numpy as np
import re
import datetime
import getpass
import traceback
from time import time
from cubical import VERSION

# This is to keep matplotlib from falling over when no DISPLAY is set (which it otherwise does,
# even if one is only trying to save figures to .png.
import matplotlib

from cubical.tools import logger
# set the base name of the logger. This must happen before any other loggers are instantiated
# (Thus before anything else that uses the logger is imported!)
logger.init("cc")

# Some modules cause issues with logging - grab their loggers and 
# manually set the log levels to something less annoying.

import logging
import warnings
logging.getLogger('matplotlib').setLevel(logging.WARNING)

GD = None

# the getuser() call fails under singularity --contain, so fall back to $USER
try:
    _user = getpass.getuser()
except:
    _user = os.environ.get("USER", "unknown")

_start_datetime = datetime.datetime.now()

_runtime_templates = dict(DATE=_start_datetime.strftime("%Y%m%d"),
                          TIME=_start_datetime.strftime("%H%M%S"),
                          USER=_user,
                          HOST=os.uname()[1],
                          ENV=os.environ)


def expand_templated_name(name, **keys):
    """
        Helper method: expands name from templated name. This uses the standard
        str.format() function, passing in GD (global dict of options), as well as any keys supplied,
        as well as the _runtime_templates dict above.
        This allows for name templates that reference both the parset, as well as runtime conditions:
        e.g. "{data[ms]}-ddid{sel[ddid]}-{DATE}-{TIME}".

        Args:
            name (str):
                the templated name
            keys (optional):
                any optional substitution keys.

        Returns:
            str:
                Expanded filename
    """
    name0 = name
    try:
        if name:
            keys.update(_runtime_templates)
            keys.update(GD)
            # substitute recursively, but up to a limit
            for i in range(10):
                name1 = name.format(**keys)
                if name1 == name:
                    break
                name = name1
        return name
    except Exception as exc:
        print("{}({})\n {}".format(type(exc).__name__, exc, traceback.format_exc()), file=log)
        if name == name0:
            print(ModColor.Str("Error substituting '{}', see above".format(name)), file=log)
        else:
            print(ModColor.Str("Error substituting '{}' (derived from '{}'), see above".format(name, name0)), file=log)
        raise ValueError(name)

from cubical.data_handler.ms_data_handler import MSDataHandler
from cubical.data_handler.wisdom import estimate_mem
from cubical.tools import parsets, dynoptparse, shm_utils, ModColor
from cubical.machines import machine_types
from cubical.machines import jones_chain_machine, jones_chain_robust_machine
from cubical.machines import ifr_gain_machine
from cubical import workers

log = logger.getLogger("main")

import cubical.solver as solver
import cubical.flagging as flagging

from cubical.statistics import SolverStats


class UserInputError(Exception):
    pass

# set to true with --Debug-Pdb 1, causes pdb to be invoked on exception
enable_pdb = True

def debug():
    """ Calls the main() function in debugging mode. """

    main(debugging=True)

def main(debugging=False):
    """
    Main cubical driver function. Reads options, sets up MS and solvers, calls the solver, etc.

    Args:
        debugging (bool, optional):
            If True, run in debugging mode.

    Raises:
        UserInputError:
            If neither --model-lsm nor --model-column were specified.
        UserInputError:
            If no Jones terms are enabled.
        UserInputError:
            If --out-mode is invalid.
        ValueError:
            If unknown Jones type is specified.
        RuntimeError:
            If I/O job on a tile failed.
    """

    # this will be set below if a custom parset is specified on the command line
    custom_parset_file = None
    # "GD" is a global defaults dict, containing options set up from parset + command line
    global GD, enable_pdb

    # keep a list of messages here, until we have a logfile open
    prelog_messages = []

    def prelog_print(level, message):
        prelog_messages.append((level, message))

    print("Using CubiCal version {}.".format(VERSION), file=log)

    try:
        if debugging:
            print("initializing from cubical.last", file=log)
            GD = pickle.load(open("cubical.last"))
            basename = GD["out"]["name"]
            parser = None
        else:
            default_parset = parsets.Parset("%s/DefaultParset.cfg" % os.path.dirname(__file__))

            # if first argument is a filename, treat it as a parset

            if len(sys.argv) > 1 and not sys.argv[1][0].startswith('-'):
                custom_parset_file = sys.argv[1]
                print("reading defaults from {}".format(custom_parset_file), file=log)
                try:
                    parset = parsets.Parset(custom_parset_file)
                except:
                    import traceback
                    traceback.print_exc()
                    raise UserInputError("'{}' must be a valid parset file. Use -h for help.".format(custom_parset_file))
                if not parset.success:
                    raise UserInputError("'{}' must be a valid parset file. Use -h for help.".format(custom_parset_file))
                # update default parameters with values from parset
                default_parset.update_values(parset, other_filename=' in {}'.format(custom_parset_file))

            import cubical
            parser = dynoptparse.DynamicOptionParser(usage='Usage: %prog [parset file] <options>',
                    description="""Questions, bug reports, suggestions: https://github.com/ratt-ru/CubiCal""",
                    version='%prog version {}'.format(cubical.VERSION),
                    defaults=default_parset.value_dict,
                    attributes=default_parset.attr_dict)

            # now read the full input from command line
            # "GD" is a global defaults dict, containing options set up from parset + command line
            GD = parser.read_input()

            # if a single argument is given, it should have been the parset
            if len(parser.get_arguments()) != (1 if custom_parset_file else 0):
                raise UserInputError("Unexpected number of arguments. Use -h for help.")

            # get dirname and basename for all output files
            outdir = expand_templated_name(GD["out"]["dir"]).strip()
            basename = expand_templated_name(GD["out"]["name"]).strip()
            can_overwrite = GD["out"]["overwrite"]
            can_backup = GD["out"]["backup"]

            explicit_basename_path = "/" in basename
            folder_is_ccout  = False

            if explicit_basename_path:
                prelog_print(0, "output basename explicitly set to {}, --out-dir setting ignored".format(basename))
                outdir = os.path.dirname(basename)
            elif outdir == "." or not outdir:
                outdir = None
                prelog_print(0, "using output basename {} in current directory".format(basename))
            else:
                # append implicit .cc-out suffix, unless already there (or ends with .cc-out)
                if not outdir.endswith("/"):
                    if outdir.endswith(".cc-out"):
                        outdir += "/"
                    else:
                        outdir += ".cc-out/"
                folder_is_ccout = outdir.endswith(".cc-out/")
                basename = outdir + basename
                if outdir != "/":
                    outdir = outdir.rstrip("/")
                prelog_print(0, "using output basename {}".format(basename))

            # create directory for output files, if specified, and it doesn't exist
            if outdir and not os.path.exists(outdir):
                prelog_print(0, "creating new output directory {}".format(outdir))
                os.mkdir(outdir)

            # are we going to be overwriting a previous run?
            out_parset = "{}.parset".format(basename)
            if os.path.exists(out_parset):
                prelog_print(0, "{} already exists, possibly from a previous run".format(out_parset))

                if can_backup:
                    if folder_is_ccout:
                        # find non-existing directory name for backup
                        backup_dir = outdir + ".0"
                        N = 0
                        while os.path.exists(backup_dir):
                            N += 1
                            backup_dir = "{}.{}".format(outdir, N)
                        # rename old directory, if we ended up manipulating the directory name
                        os.rename(outdir, backup_dir)
                        os.mkdir(outdir)
                        prelog_print(0, ModColor.Str("backed up existing {} to {}".format(outdir, backup_dir), "blue"))
                    else:
                        prelog_print(0, "refusing to auto-backup output directory, since it is not a .cc-out dir")

                if os.path.exists(out_parset):
                    if can_overwrite:
                        prelog_print(0, "proceeding anyway since --out-overwrite is set")
                    else:
                        if folder_is_ccout:
                            prelog_print(0, "won't proceed without --out-overwrite and/or --out-backup")
                        else:
                            prelog_print(0, "won't proceed without --out-overwrite")
                        raise UserInputError("{} already exists: won't overwrite previous run".format(out_parset))

            GD["out"]["name"] = basename

            # "GD" is a global defaults dict, containing options set up from parset + command line
            pickle.dump(GD, open("cubical.last", "wb"))

            # save parset with all settings
            parser.write_to_parset(out_parset)

        enable_pdb = GD["debug"]["pdb"]

        if GD["debug"]["escalate-warnings"]:
            warnings.simplefilter('error', UserWarning)
            np.seterr(all='raise')
            if GD["debug"]["escalate-warnings"] > 1:
                warnings.simplefilter('error', Warning)
                log(0).print("all warnings will be escalated to exceptions")
            else:
                log(0).print("UserWarnings will be escalated to exceptions")

        # clean up shared memory from any previous runs
        shm_utils.cleanupStaleShm()

        # now setup logging
        logger.logToFile(basename + ".log", append=GD["log"]["append"])
        logger.enableMemoryLogging(GD["log"]["memory"])
        logger.setBoring(GD["log"]["boring"])
        logger.setGlobalVerbosity(GD["log"]["verbose"])
        logger.setGlobalLogVerbosity(GD["log"]["file-verbose"])

        if not debugging:
            print("started " + " ".join(sys.argv), file=log)

        # dump accumulated messages from before log was open
        for level, message in prelog_messages:
            print(message, file=log(level))
        prelog_messages = []

        # clean up shared memory from any previous runs
        shm_utils.cleanupStaleShm()

        # disable matplotlib's tk backend if we're not going to be showing plots
        if GD['out']['plots'] =='show' or GD['madmax']['plot'] == 'show':
            import pylab
            try:
                pylab.figure()
                pylab.close()
            except Exception as exc:
                import traceback
                print(ModColor.Str("Error initializing matplotlib: {}({})\n {}".format(type(exc).__name__,
                                                                                       exc, traceback.format_exc())), file=log)
                raise UserInputError("matplotlib can't connect to X11. Can't use --out-plots show or --madmax-plot show.")
        else:
            matplotlib.use("Agg")

        # print current options
        if parser is not None:
            parser.print_config(dest=log)

        double_precision = GD["sol"]["precision"] == 64

        # set up RIME

        solver_opts = GD["sol"]
        debug_opts  = GD["debug"]
        out_opts = GD["out"]
        sol_jones = solver_opts["jones"]
        if isinstance(sol_jones, string_types):
            sol_jones = set(sol_jones.split(','))
        jones_opts = [GD[j.lower()] for j in sol_jones]
        # collect list of options from enabled Jones matrices
        if not len(jones_opts):
            raise UserInputError("No Jones terms are enabled")
        print(ModColor.Str("Enabling {}-Jones".format(",".join(sol_jones)), col="green"), file=log)

        have_dd_jones = any([jo['dd-term'] for jo in jones_opts])
        have_solvables = any([jo['solvable'] for jo in jones_opts])

        solver.GD = GD

        # set up data handler

        solver_type = GD['out']['mode']
        if solver_type not in solver.SOLVERS:
            raise UserInputError("invalid setting --out-mode {}".format(solver_type))
        solver_mode_name = solver.SOLVERS[solver_type].__name__.replace("_", " ")
        print(ModColor.Str("mode: {}".format(solver_mode_name), col='green'), file=log)
        # these flags are used below to tweak the behaviour of gain machines and model loaders
        apply_only = solver.SOLVERS[solver_type].is_apply_only
        print("solver is apply-only type: {}".format(apply_only), file=log(0))
        load_model = solver.SOLVERS[solver_type].is_model_required
        print("solver requires model: {}".format(load_model), file=log(0))
        
        if not apply_only and not have_solvables:
            raise UserInputError("No Jones terms have been marked as solvable")

        if load_model and not GD["model"]["list"]:
            raise UserInputError("--model-list must be specified")

        ms = MSDataHandler(GD["data"]["ms"],
                           GD["data"]["column"],
                           output_column=GD["out"]["column"],
                           output_model_column=GD["out"]["model-column"],
                           output_weight_column=GD["out"]["weight-column"],
                           reinit_output_column=GD["out"]["reinit-column"],
                           taql=GD["sel"]["taql"],
                           fid=GD["sel"]["field"],
                           ddid=GD["sel"]["ddid"],
                           channels=GD["sel"]["chan"],
                           diag=GD["sel"]["diag"],
                           beam_pattern=GD["model"]["beam-pattern"],
                           beam_l_axis=GD["model"]["beam-l-axis"],
                           beam_m_axis=GD["model"]["beam-m-axis"],
                           active_subset=GD["sol"]["subset"] if not apply_only else None,
                           min_baseline=GD["sol"]["min-bl"] if not apply_only else 0,
                           max_baseline=GD["sol"]["max-bl"] if not apply_only else 0,
                           chunk_freq=GD["data"]["freq-chunk"],
                           rebin_freq=GD["data"]["rebin-freq"],
                           do_load_CASA_kwtables = GD["out"]["casa-gaintables"],
                           feed_rotate_model=GD["model"]["feed-rotate"],
                           pa_rotate_model=GD["model"]["pa-rotate"],
                           pa_rotate_montblanc=GD["montblanc"]["pa-rotate"],
                           derotate_output=GD["out"]["derotate"],
                           do_normalize_data=GD["data"]["normalize"]
                           )

        solver.metadata = ms.metadata
        # if using dual-corr mode, propagate this into Jones options
        if ms.ncorr == 2:
            for jo in jones_opts:
                jo['diag-only'] = True
                jo['diag-data'] = True
            solver_opts['diag-only'] = True
            solver_opts['diag-data'] = True

        # With a single Jones term, create a gain machine factory based on its type.
        # With multiple Jones, create a ChainMachine factory
        term_iters = solver_opts["term-iters"]
        if type(term_iters) is int:
            term_iters = [term_iters] * len(jones_opts)
            solver_opts["term-iters"] = term_iters
            len(jones_opts) > 1 and log.warn("Multiple gain terms specified, but a recipe of solver sol-term-iters not given. "
                                             "This may indicate user error. We will assume doing the same number of iterations per term and "
                                             "stopping on the last term on the chain.")
        elif type(term_iters) is list and len(term_iters) == 1:
            term_iters = term_iters * len(jones_opts)
            solver_opts["term-iters"] = term_iters
            len(jones_opts) > 1 and log.warn("Multiple gain terms specified, but a recipe of solver sol-term-iters not given. "
                                             "This may indicate user error. We will assume doing the same number of iterations per term and "
                                             "stopping on the last term on the chain.")
        elif type(term_iters) is list and len(term_iters) < len(jones_opts):
            raise ValueError("sol-term-iters is a list, but does not match or exceed the number of gain terms being solved. "
                             "Please either only set a single value to be used or provide a list to construct a iteration recipe")
        elif type(term_iters) is list and len(term_iters) >= len(jones_opts):
            pass # user is executing a recipe
        else:
            raise TypeError("sol-term-iters is neither a list, nor a int. Check your parset")

        if len(jones_opts) == 1:
            jones_opts = jones_opts[0]
            # for just one term, propagate --sol-term-iters, if set, into its max-iter setting
            term_iters = solver_opts["term-iters"]
            if term_iters:
                jones_opts["max-iter"] = term_iters[0] if hasattr(term_iters,'__getitem__') else term_iters
            # create a gain machine factory
            jones_class = machine_types.get_machine_class(jones_opts['type'])
            if jones_class is None:
                raise UserInputError("unknown Jones type '{}'".format(jones_opts['type']))
        elif jones_opts[0]['type'].startswith("robust"):
            jones_class = jones_chain_robust_machine.JonesChain
        else:
            jones_class = jones_chain_machine.JonesChain

        # init models
        dde_mode = GD["model"]["ddes"]

        if dde_mode == 'always' and not have_dd_jones:
            raise UserInputError("we have '--model-ddes always', but no direction dependent Jones terms enabled")

        # force floats in Montblanc calculations
        mb_opts = GD["montblanc"]
        # mb`_opts['dtype'] = 'float'

        ms.init_models(str(GD["model"]["list"]).split(","),
                       GD["weight"]["column"].split(",") if GD["weight"]["column"] else None,
                       fill_offdiag_weights=GD["weight"]["fill-offdiag"],
                       mb_opts=GD["montblanc"],
                       use_ddes=have_dd_jones and dde_mode != 'never',
                       degrid_opts=GD["degridding"])

        if len(ms.model_directions) < 2 and have_dd_jones and dde_mode == 'auto':
            raise UserInputError("--model-list does not specify directions. "
                    "Have you forgotten a @dE tag perhaps? Rerun with '--model-ddes never' to proceed anyway.")

        if load_model:
            # set up subtraction options
            solver_opts["subtract-model"] = smod = GD["out"]["subtract-model"]
            if smod < 0 or smod >= len(ms.models):
                raise UserInputError("--out-subtract-model {} out of range for {} model(s)".format(smod, len(ms.models)))

            # parse subtraction directions as a slice or list
            subdirs = GD["out"]["subtract-dirs"]
            if type(subdirs) is int:
                subdirs = [subdirs]
            if subdirs:
                if isinstance(subdirs, string_types):
                    try:
                        if ',' in subdirs:
                            subdirs = list(map(int, subdirs.split(",")))
                        else:
                            subdirs = eval("np.s_[{}]".format(subdirs))
                    except:
                        raise UserInputError("invalid --out-subtract-model option '{}'".format(subdirs))
                elif type(subdirs) is not list:
                    raise UserInputError("invalid --out-subtract-dirs option '{}'".format(subdirs))
                # check ranges
                if type(subdirs) is list:
                    out_of_range = [ d for d in subdirs if d < 0 or d >= len(ms.model_directions) ]
                    if out_of_range:
                        raise UserInputError("--out-subtract-dirs {} out of range for {} model direction(s)".format(
                                ",".join(map(str, out_of_range)), len(ms.model_directions)))
                print("subtraction directions set to {}".format(subdirs), file=log(0))
            else:
                subdirs = slice(None)
            solver_opts["subtract-dirs"] = subdirs

        # create gain machine factory
        # TODO: pass in proper antenna and correlation names, rather than number
        solver_opts["correct-dir"] = GD["out"]["correct-dir"] if GD["out"]["correct-dir"] >= 0 else None

        grid = dict(dir=list(range(len(ms.model_directions) or 1)), ant=ms.antnames, corr=list(ms.feeds), time=ms.uniq_times, freq=ms.all_freqs)
        solver.gm_factory = jones_class.create_factory(grid=grid,
                                                       apply_only=apply_only,
                                                       double_precision=double_precision,
                                                       global_options=GD, jones_options=jones_opts)

        solver.gm_factory.set_metadata(ms)
                                                       
        # create IFR-based gain machine. Only compute gains if we're loading a model
        # (i.e. not in load-apply mode)
        solver.ifrgain_machine = ifr_gain_machine.IfrGainMachine(solver.gm_factory, GD["bbc"], compute=load_model)

        solver.legacy_version12_weights = GD["weight"]["legacy-v1-2"]

        single_chunk = GD["data"]["single-chunk"]
        single_tile = GD["data"]["single-tile"]


        # set up chunking

        chunk_by = GD["data"]["chunk-by"]
        if isinstance(chunk_by, string_types):
            chunk_by = chunk_by.split(",")
        jump = float(GD["data"]["chunk-by-jump"])

        if single_chunk:
            chunks_per_tile = 1
            max_chunks_per_tile = 1
        else:
            nw = GD["dist"]["nworker"] or ((GD["dist"]["ncpu"] / (GD["dist"]["nthread"] or 1)) or 1) - 1
            chunks_per_tile = max(GD["dist"]["min-chunks"], nw, 1)
            max_chunks_per_tile = 0
            if GD["dist"]["max-chunks"]:
                chunks_per_tile = max(max(GD["dist"]["max-chunks"], chunks_per_tile), 1)
                max_chunks_per_tile = GD["dist"]["max-chunks"]

        print("defining chunks (time {}, freq {}{})".format(GD["data"]["time-chunk"], GD["data"]["freq-chunk"],
            ", also when {} jumps > {}".format(", ".join(chunk_by), jump) if chunk_by else ""), file=log)

        chunks_per_tile, tile_list = ms.define_chunk(GD["data"]["time-chunk"], GD["data"]["rebin-time"],
                                            GD["data"]["freq-chunk"],
                                            chunk_by=chunk_by, chunk_by_jump=jump,
                                            chunks_per_tile=chunks_per_tile, max_chunks_per_tile=max_chunks_per_tile)

        # setup worker process properties
        workers.setup_parallelism(GD["dist"]["ncpu"], GD["dist"]["nworker"], GD["dist"]["nthread"],
                                  debugging or single_chunk,
                                  GD["dist"]["pin"], GD["dist"]["pin-io"], GD["dist"]["pin-main"],
                                  ms.use_montblanc, GD["montblanc"]["threads"],
                                  max_workers=chunks_per_tile)


        # Estimate memory usage. This is still experimental.
        estimate_mem(ms, tile_list, GD["data"], GD["dist"])

        # now that we have tiles, define the flagging situation (since this may involve a one-off iteration through the
        # MS to populate the column)
        ms.define_flags(tile_list, flagopts=GD["flags"])

        # single-chunk implies single-tile
        if single_tile >= 0:
            tile_list = tile_list[single_tile:single_tile+1]
            print("--data-single-tile {} set, will process only the one tile".format(single_tile), file=log(0, "blue"))
        elif single_chunk:
            match = re.match("D([0-9]+)T([0-9]+)", single_chunk)
            if not match:
                raise ValueError("invalid setting: --data-single-chunk {}".format(single_chunk))
            ddid_tchunk = int(match.group(1)), int(match.group(2))

            tilemap = { (rc.ddid, rc.tchunk): (tile, rc) for tile in tile_list for rc in tile.rowchunks }
            single_tile_rc = tilemap.get(ddid_tchunk)
            if single_tile_rc:
                tile, rc = single_tile_rc
                tile_list = [tile]
                print("--data-single-chunk {} in {}, rows {}:{}".format(
                    single_chunk, tile.label, min(rc.rows0), max(rc.rows0)+1), file=log(0, "blue"))
            else:
                raise ValueError("--data-single-chunk {}: chunk with this ID not found".format(single_chunk))

        # run the main loop

        t0 = time()

        stats_dict = workers.run_process_loop(ms, tile_list, load_model, single_chunk, solver_type, solver_opts, debug_opts, out_opts)


        print(ModColor.Str("Time taken for {}: {} seconds".format(solver_mode_name, time() - t0), col="green"), file=log)

        # print flagging stats
        print(ModColor.Str("Flagging stats: ",col="green") + " ".join(ms.get_flag_counts()), file=log)

        if not apply_only:
            # now summarize the stats
            print("computing summary statistics", file=log)
            st = SolverStats(stats_dict)
            filename = basename + ".stats.pickle"
            st.save(filename)
            print("saved summary statistics to %s" % filename, file=log)
            print_stats = GD["log"]["stats"]
            if print_stats:
                print("printing some summary statistics below", file=log(0))
                thresholds = []
                for thr in GD["log"]["stats-warn"].split(","):
                    field, value = thr.split(":")
                    thresholds.append((field, float(value)))
                    print("  highlighting {}>{}".format(field, float(value)), file=log(0))
                if print_stats == "all":
                    print_stats = st.get_notrivial_chunk_statfields()
                else:
                    print_stats = print_stats.split("//")
                for stats in print_stats:
                    if stats[0] != "{":
                        stats = "{{{}}}".format(stats)
                    lines = st.format_chunk_stats(stats, threshold=thresholds)
                    print("  summary stats for {}:\n  {}".format(stats, "\n  ".join(lines)), file=log(0))

            if GD["postmortem"]["enable"]:
                # flag based on summary stats
                flag3 = flagging.flag_chisq(st, GD, basename, ms.nddid_actual)

                if flag3 is not None:
                    st.apply_flagcube(flag3)
                    if GD["flags"]["save"] and flag3.any() and not GD["data"]["single-chunk"]:
                        print("regenerating output flags based on post-solution flagging", file=log)
                        flagcol = ms.flag3_to_col(flag3)
                        ms.save_flags(flagcol)

            # make plots
            if GD["out"]["plots"]:
                import cubical.plots
                try:
                    cubical.plots.make_summary_plots(st, ms, GD, basename)
                except Exception as exc:
                    if GD["debug"]["escalate-warnings"]:
                        raise
                    import traceback
                    print(file=ModColor.Str("An error has occurred while making summary plots: {}({})\n {}".format(type(exc).__name__,
                                                                                           exc,
                                                                                           traceback.format_exc())))
                    print(ModColor.Str("This is not fatal, but should be reported (and your plots have gone missing!)"), file=log)

        # make BBC plots
        if solver.ifrgain_machine and solver.ifrgain_machine.is_computing() and GD["bbc"]["plot"] and GD["out"]["plots"]:
            import cubical.plots.ifrgains
            if GD["debug"]["escalate-warnings"]:
                with warnings.catch_warnings():
                    warnings.simplefilter("error", np.ComplexWarning)
                    cubical.plots.ifrgains.make_ifrgain_plots(solver.ifrgain_machine.reload(), ms, GD, basename)
            else:
                try:
                    cubical.plots.ifrgains.make_ifrgain_plots(solver.ifrgain_machine.reload(), ms, GD, basename)
                except Exception as exc:
                    import traceback
                    print(file=ModColor.Str("An error has occurred while making BBC plots: {}({})\n {}".format(type(exc).__name__,
                                                                                           exc,
                                                                                           traceback.format_exc())))
                    print(ModColor.Str("This is not fatal, but should be reported (and your plots have gone missing!)"), file=log)

        ms.close()

        print(ModColor.Str("completed successfully", col="green"), file=log)

    except RuntimeWarning:
        from cubical.tools import pdb
        pdb.set_trace()

    except Exception as exc:
        for level, message in prelog_messages:
            print(message, file=log(level))

        if type(exc) is UserInputError:
            print(ModColor.Str(exc), file=log)
        else:
            import traceback
            print(ModColor.Str("Exiting with exception: {}({})\n {}".format(type(exc).__name__,
                                                                    exc, traceback.format_exc())), file=log)
            if enable_pdb and type(exc) is not UserInputError:
                warnings.filterwarnings("default")  # in case pdb itself throws a warning
                from cubical.tools import pdb
                exc, value, tb = sys.exc_info()
                pdb.post_mortem(tb)
        sys.exit(2 if type(exc) is UserInputError else 1)



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Kernels for the Jones chain machine. Functions require output arrays to be 
provided. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels
from cubical.kernels import generics
from cubical.kernels import full_complex

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

# Allocators are the same as for the 2x2 complex kernel.
allocate_vis_array = full_complex.allocate_vis_array
allocate_gain_array = full_complex.allocate_gain_array
allocate_flag_array = full_complex.allocate_flag_array

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jh(jh, g, t_int, f_int):
    """
    Given J\ :sup:`H` (initially populated with the model array) and gains, computes the non-zero 
    elements of J\ :sup:`H`. J\ :sup:`H` has full time and frequency resolution - solution intervals
    are used to correctly associate the gains with the model. The result here contains the useful 
    elements of J\ :sup:`H` but does not look like the analytic solution. This function is called
    multiple times during a chain computation. Each call applies a different gain term.   

    Args:
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimension (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = jh.shape[0]
    n_mod = jh.shape[1]
    n_tim = jh.shape[2]
    n_fre = jh.shape[3]
    n_ant = jh.shape[4]
    g_dir = g.shape[0]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j])
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl,0], all_bls[ibl,1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]

                        jh00 = jh[d,i,t,f,aa,ab,0,0]
                        jh10 = jh[d,i,t,f,aa,ab,1,0]
                        jh01 = jh[d,i,t,f,aa,ab,0,1]
                        jh11 = jh[d,i,t,f,aa,ab,1,1]

                        g00 = g[bd,bt,bf,aa,0,0]
                        g01 = g[bd,bt,bf,aa,0,1]
                        g10 = g[bd,bt,bf,aa,1,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        jh[d,i,t,f,aa,ab,0,0] = (g00*jh00 + g01*jh10)
                        jh[d,i,t,f,aa,ab,0,1] = (g00*jh01 + g01*jh11)
                        jh[d,i,t,f,aa,ab,1,0] = (g10*jh00 + g11*jh10)
                        jh[d,i,t,f,aa,ab,1,1] = (g10*jh01 + g11*jh11)

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def apply_left_inv_jones(jhr, ginv, t_int, f_int):
    """
    Applies the inverse of a gain array the left side of J\ :sup:`H`\R. J\ :sup:`H`\R has full time 
    and frequency resolution - solution intervals are used to correctly associate the gains with 
    the model. This is a special function which is unique to the Jones chain.    

    Args:
        jhr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, t, f, a, c, c).
        ginv (np.complex64 or np.complex128):
            Typed memoryview of inverse gain array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = jhr.shape[0]
    n_tim = jhr.shape[1]
    n_fre = jhr.shape[2]
    n_ant = jhr.shape[3]

    g_dir = ginv.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):
                    bd = broadcast_dirs[d]

                    jhr00 = jhr[d,t,f,aa,0,0]
                    jhr01 = jhr[d,t,f,aa,0,1]
                    jhr10 = jhr[d,t,f,aa,1,0]
                    jhr11 = jhr[d,t,f,aa,1,1]

                    ginv00 = ginv[bd,bt,bf,aa,0,0]
                    ginv01 = ginv[bd,bt,bf,aa,0,1]
                    ginv10 = ginv[bd,bt,bf,aa,1,0]
                    ginv11 = ginv[bd,bt,bf,aa,1,1]

                    jhr[d,t,f,aa,0,0] = (ginv00*jhr00 + ginv01*jhr10)
                    jhr[d,t,f,aa,0,1] = (ginv00*jhr01 + ginv01*jhr11)
                    jhr[d,t,f,aa,1,0] = (ginv10*jhr00 + ginv11*jhr10)
                    jhr[d,t,f,aa,1,1] = (ginv10*jhr01 + ginv11*jhr11)

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def sum_jhr_intervals(jhr, jhrint, t_int, f_int):
    """
    Collapses J\ :sup:`H`\R to be cosistent with the solution intervals for the current gain of 
    interest. This is necessary as each gain term in a chain may have unique solution intervals. 

    Args:
        jhr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, t, f, a, c, c).
        jhrint (np.complex64 or np.complex128):
            Typed memoryview of collapsed J\ :sup:`H`\R array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = jhr.shape[0]
    n_tim = jhr.shape[1]
    n_fre = jhr.shape[2]
    n_ant = jhr.shape[3]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):

                    jhrint[d,bt,bf,aa,0,0] += jhr[d,t,f,aa,0,0]
                    jhrint[d,bt,bf,aa,0,1] += jhr[d,t,f,aa,0,1]
                    jhrint[d,bt,bf,aa,1,0] += jhr[d,t,f,aa,1,0]
                    jhrint[d,bt,bf,aa,1,1] += jhr[d,t,f,aa,1,1]

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_residual(m, r):
    """
    Given the model array (already multipled by the gains) and the residual array (already populated
    with the observed data), computes the final residual. This is a special case where the gains
    are applied to the model elsewhere.

    NOTE: It may be possible to accelerate this if m and r and already correctly populated in the
    lower triangle. Then we can loop over all baselines and do the subtraction everywhere instead
    of the conjugation.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of the model array with dimensions (d, m, t, f, a, a, c, c).
        r (np.complex64 or np.complex128):
            Typed memoryview of tje residual array with dimensions (m, t, f, a, a, c, c).
    """

    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i+1, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]  
        for i in range(n_mod):
            for t in range(n_tim):
                for f in range(n_fre):
                    for d in range(n_dir):

                        r[i,t,f,aa,ab,0,0] -= m[d,i,t,f,aa,ab,0,0]
                        r[i,t,f,aa,ab,0,1] -= m[d,i,t,f,aa,ab,0,1]
                        r[i,t,f,aa,ab,1,0] -= m[d,i,t,f,aa,ab,1,0]
                        r[i,t,f,aa,ab,1,1] -= m[d,i,t,f,aa,ab,1,1]

                        r[i,t,f,ab,aa,0,0] = r[i,t,f,aa,ab,0,0].conjugate()
                        r[i,t,f,ab,aa,1,0] = r[i,t,f,aa,ab,0,1].conjugate()
                        r[i,t,f,ab,aa,0,1] = r[i,t,f,aa,ab,1,0].conjugate()
                        r[i,t,f,ab,aa,1,1] = r[i,t,f,aa,ab,1,1].conjugate()


# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Cython kernels for the diagonal vis data phase-only gain machine. Functions require output arrays to be
provided. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels
from cubical.kernels import generics
from cubical.kernels import full_complex
from cubical.kernels import diag_complex
from cubical.kernels import diagdiag_complex
from cubical.kernels import phase_only

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

# Allocators same as for generic full kernel
allocate_vis_array = full_complex.allocate_vis_array
allocate_gain_array = full_complex.allocate_gain_array
allocate_flag_array = full_complex.allocate_flag_array
allocate_param_array = full_complex.allocate_param_array

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhj(m, jhj, t_int=1, f_int=1):
    """
    Given the model array, computes the diagonal entries of J\ :sup:`H`\J. J\ :sup:`H`\J is computed
    over intervals. This is an approximation of the Hessian. In the phase-only case, this
    approximation does not depend on the gains, therefore it does not vary with iteration.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model array with dimensions (d, m, t, f, a, a, c, c).
        jhj (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\J array with dimensions (d, ti, fi, a, c, c). Must be zero-filled.
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j])
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl,0], all_bls[ibl,1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]

                        jhj[d,bt,bf,aa,0,0] += (m00*m00.conjugate())
                        jhj[d,bt,bf,aa,1,1] += (m11*m11.conjugate())

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhr(gh, jh, r, jhr, t_int=1, f_int=1):
    """
    Given the conjugate gains, J\ :sup:`H` and the residual (or observed data, in special cases),
    computes J\ :sup:`H`\R. J\ :sup:`H`\R is computed over intervals.

    Args:
        gh (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimension (d, ti, fi, a, c, c).
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        jhr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = jh.shape[0]
    n_mod = jh.shape[1]
    n_tim = jh.shape[2]
    n_fre = jh.shape[3]
    n_ant = jh.shape[4]
    g_dir = gh.shape[0]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j])
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl,0], all_bls[ibl,1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]

                        r00 = r[i,t,f,aa,ab,0,0]
                        r11 = r[i,t,f,aa,ab,1,1]

                        jhh00 = jh[d,i,t,f,ab,aa,0,0]
                        jhh11 = jh[d,i,t,f,ab,aa,1,1]

                        jhr[d,bt,bf,aa,0,0] += gh[bd,bt,bf,aa,0,0]*r00*jhh00
                        jhr[d,bt,bf,aa,1,1] += gh[bd,bt,bf,aa,1,1]*r11*jhh11

# Remaining kernel functions are reused.

# J^H is computed assuming diagonal gains.
compute_jh = diagdiag_complex.compute_jh

# J^H.J inverse is computed assuming diagonal blocks.
compute_jhjinv = generics.compute_diag_inverse

# Residuals computed assuming diagonal gains.
compute_residual = diagdiag_complex.compute_residual

# Corrected visibilities computed assuming diagonal gains.
compute_corrected = diagdiag_complex.compute_corrected

# Gains applied as diagonal.
apply_gains = diagdiag_complex.apply_gains

# Update computed as in the general phase-only case.
compute_update = phase_only.compute_update

# # Gains inverted as diagonal.
# invert_gains = generics.compute_diag_inverse



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
import numpy as np
from numpy.ma import masked_array

from cubical import param_db

from cubical.tools import logger, ModColor
log = logger.getLogger("gain_machine")


class IfrGainMachine(object):
    """
    Interferometer-based gain machine.
    
    Note that this class is outside of the normal gain machine hierarchy. It wraps all the 
    functionality required to manage IFR-based gain (aka baseline-based corrections) 
    calculations.
    """

    def __init__(self, gmfactory, ifrgain_opts, compute=True):
        """
        Initializes the IFR-based gains machinery.
        
        Args:
            gmfactory:      a GainMachine Factory is used to manage the solution databases
            ifrgain_opts:   dict of options
            compute:        if False, gains are not computed even if options ask them to
        """
        from cubical.main import expand_templated_name
        self.gmfactory = gmfactory
        load_from = expand_templated_name(ifrgain_opts['load-from'])
        save_to = expand_templated_name(ifrgain_opts['save-to'])
        self._ifrgains_per_chan = ifrgain_opts['per-chan']
        self._ifrgain = None
        self._nfreq = gmfactory.grid["freq"]
        nfreq, nant, ncorr = [len(gmfactory.grid[axis]) for axis in ("freq", "ant", "corr")]
        if load_from:
            filename = load_from
            print(ModColor.Str("applying baseline-based corrections (BBCs) from {}".format(filename),
                                         col="green"), file=log(0))
            if "//" in filename:
                filename, prefix = filename.rsplit("//", 1)
            else:
                filename, prefix = filename, "BBC"
            parm = param_db.load(filename).get(prefix)
            if parm is None:
                print(ModColor.Str("  no solutions for '{}' in {}".format(prefix, filename)), file=log(0))
            else:
                self._ifrgain = parm.reinterpolate(freq=gmfactory.grid["freq"]).filled()
                if tuple(self._ifrgain.shape) != (nfreq, nant, nant, ncorr, ncorr):
                    print(ModColor.Str("  invalid BBC shape {}, will ignore".format(self._ifrgain.shape)), file=log(0))
                    self._ifrgain = None
                else:
                    print("  loaded per-channel BBCs of shape {}".format(filename, self._ifrgain.shape), file=log(0))
                    if not self._ifrgains_per_chan:
                        print("  using one mean value across band", file=log(0))
                        self._ifrgain[np.newaxis,...] = self._ifrgain.mean(axis=0)
                    # reset off-diagonal values, if needed
                    if ifrgain_opts["apply-2x2"]:
                        print(ModColor.Str(
                            "  using full 2x2 BBCs. You'd better know what you're doing!",col="green"), file=log(0))
                    else:
                        self._ifrgain[..., (0, 1), (1, 0)] = 1
                        print("  using parallel-hand BBCs only", file=log(0))
        if save_to and compute:
            self._compute_2x2 = ifrgain_opts["compute-2x2"]
            # setup axes for IFR-based gains
            axes = ["freq", "ant1", "ant2", "corr1", "corr2"]
            # define the ifrgain parameter
            self._save_filename = save_to
            parm = gmfactory.define_param(self._save_filename, "BBC", 1+0j, axes, interpolation_axes=["freq"])
            self._ifrgains_grid = {axis: parm.grid[i] for i, axis in enumerate(axes)}
            # initialize accumulators for M.D^H and D.D^H terms
            self._mdh_sum = np.ma.zeros(parm.shape, gmfactory.ctype, fill_value=0)
            self._ddh_sum = np.ma.zeros(parm.shape, gmfactory.ctype, fill_value=0)
            #
            print("will compute & save suggested baseline-based corrections (BBCs) to {}".format(
                            self._save_filename), file=log(0))
            print("  (these can optionally be applied in a subsequent CubiCal run)", file=log(0))
        else:
            self._ifrgains_grid = None

    def is_applying(self):
        """Returns True if we have valid IFR gains to apply"""
        return self._ifrgain is not None

    def is_computing(self):
        """Returns True if we're computing output IFR gains"""
        return self._ifrgains_grid is not None

    def apply(self, obser_arr, freq_slice):
        """
        Applies IFR-based gains (if any) to input data
        
        Args:
            obser_arr:   data array (shape ntime,nfreq,nant,nant,ncorr,ncorr)
            freq_slice: frequency slice in data array
        """
        # IFR gains are freq,ant,ant,corr,corr, so broadcast the time axes, and slice the freq axis
        if self._ifrgain is not None:
            obser_arr *= self._ifrgain[np.newaxis, freq_slice, ...]

    def update(self, obser_arr, corrupt_model, flags, freq_slice, soldict):
        """
        Computes update to interferometer gains based on data and corrupt model. Saves this to 
        SharedDict of solutions.

        Args:
            obser_arr:      data array (nmod,ntime,nfreq,nant,nant,ncorr,ncorr) 
            corrupt_model:  corrupt model array (nmod,ntime,nfreq,nant,nant,ncorr,ncorr)
            flags:          flags array, same shape as data
            freq_slice:     frequency slice which is being updated
            soldict:        shared dict to save gain update to
        """
        if not self.is_computing():
            return
        # shape of BB arrays is nfreq,nant,nant,ncorr,ncorr or nant,nant,ncorr,ncorr
        bbshape = obser_arr.shape[-5:]
        DDH = soldict.addSharedArray('ifrgains:DDH__', bbshape, dtype=obser_arr.dtype)
        MDH = soldict.addSharedArray('ifrgains:MDH__', bbshape, dtype=obser_arr.dtype)
        fl = soldict.addSharedArray('ifrgains:flags__', bbshape[:-2], dtype=bool)
        soldict["ifrgains:freqslice__"] = freq_slice

        D, M = obser_arr, corrupt_model
        DH = np.conj(D)
        DDH[:] = (D * DH).sum(axis=(0,1))
        MDH[:] = (M * DH).sum(axis=(0,1))
        fl[:]  = flags.all(axis=0)

    def accumulate(self, subdict):
        """
        Gets solutions from SharedDict populated by update(), and updates internal accumulators
        based on this. 
        Args:
            subdict: dict populated by update() above

        Returns:

        """
        # update ifr gains
        if not self.is_computing():
            return
        ddh = subdict['ifrgains:DDH__']
        # promote corr1,corr2 axes in flags, to make them into a proper mask
        fl = np.empty(ddh.shape, bool)
        fl[:] = subdict['ifrgains:flags__'][...,np.newaxis,np.newaxis]
        ddh = masked_array(ddh, fl, fill_value=0)
        mdh = masked_array(subdict['ifrgains:MDH__'], fl, fill_value=0)
        self._ddh_sum[subdict['ifrgains:freqslice__'],...] += ddh
        self._mdh_sum[subdict['ifrgains:freqslice__'],...] += mdh

    def save(self):
        """
        Finalizes the IFR gain solutions, and saves them to the database  
        """
        if not self.is_computing():
            return
        ## if not per-channel, then sum over channels
        if not self._ifrgains_per_chan:
            self._ddh_sum = self._ddh_sum.sum(0)
            self._mdh_sum = self._mdh_sum.sum(0)

        # mask out null values
        self._ddh_sum.mask |= (self._ddh_sum==0)
        with np.errstate(under='ignore'): # work around https://github.com/numpy/numpy/issues/4895
            ifrgain = self._mdh_sum/self._ddh_sum
        ifrgain.data[self._ddh_sum.mask] = 1
        ifrgain.fill_value = 1

        # promote freq axis back out
        if not self._ifrgains_per_chan:
            ifrgain = np.ma.vstack([ifrgain]*len(self._nfreq))

        # multiply by previous IFR gain, if that was set. This will promote a frequency axis
        if self._ifrgain is not None:
            ifrgain *= self._ifrgain

        # if not 2x2, then reset off-diagonals
        if not self._compute_2x2:
            ifrgain[...,0,1] = ifrgain[...,1,0] = 1

        self.gmfactory.get_solution_db("BBC").add_chunk("BBC", ifrgain, self._ifrgains_grid)

    def reload(self):
        """Reloads saved IFR gain solutions from database"""
        return param_db.load(self._save_filename)["BBC"].get_cube()


from __future__ import print_function
from builtins import range
import numpy as np
import traceback

from cubical.tools import BREAK  # useful: can set static breakpoints by putting BREAK() in the code

from cubical.tools import logger
log = logger.getLogger("madmax")

def make_dual_absres_plot(absres, fl_prior, fl_new, p, q, metadata, subplot_titles={}):
    import pylab
    feeds = metadata.feeds
    # inv: data that was flagged prior to this mad max step
    fl_prior = fl_prior[:, :, p, q] & ~fl_new[:, :, p, q]
    figure = pylab.figure(figsize=(16, 10))
    resmask = np.zeros_like(absres[0, :, :, p, q], dtype=bool)
    resmask[:] = fl_prior[..., np.newaxis, np.newaxis] #| absres[0, :, :, p, q]==0
    res = np.ma.masked_array(absres[0, :, :, p, q], resmask)
    resmin, resmax = res.min(), res.max()
    vmin = float(max(resmin if resmin is not np.ma.masked else 0, 1e-9))
    vmax = float(resmax if resmax is not np.ma.masked else 1)
    if vmin != vmax:
        from matplotlib.colors import LogNorm
        norm = LogNorm(vmin, vmax)
    else:
        norm = None
    for c1, x1 in enumerate(feeds.upper()):
        for c2, x2 in enumerate(feeds.upper()):
            pylab.subplot(2, 4, 1 + c1 * 2 + c2)
            if (res[..., c1, c2] > 0).any():
                pylab.imshow(res[..., c1, c2], norm=norm, aspect='auto')
                try:
                    pylab.colorbar()
                except ZeroDivisionError:
                    pass
            pylab.title(subplot_titles.get((c1,c2)))
    for c1, x1 in enumerate(feeds.upper()):
        for c2, x2 in enumerate(feeds.upper()):
            pylab.subplot(2, 4, 5 + c1 * 2 + c2)
            res1 = np.ma.masked_array(absres[0, :, :, p, q, c1, c2], fl_prior | fl_new[:, :, p, q])
            if (res1 > 0).any():
                pylab.imshow(res1, norm=norm, aspect='auto')
                try:
                    pylab.colorbar()
                except ZeroDivisionError:
                    pass
            pylab.title("{}{} flagged".format(x1, x2))
    return figure



def make_baseline_mad_plot(mad, medmad, med_thr, metadata, max_label="", chunk_label="", antenna_mad_threshold=0):
    import pylab
    colors = [["black", "red"], ["green", "blue"]]
    n_mod = mad.shape[0]
    n_ant = mad.shape[1]
    per_corr = medmad.ndim == 3
    n_cor = medmad.shape[1]

    outflags = np.zeros(mad.shape[1:],bool)

    # sort baselines by length and form up index list
    baselines = [ (p,q) for p in range(n_ant) for q in range(p+1, n_ant) ]
    # sort arrays by baseline length
    indices_pq = sorted([(metadata.baseline_length[p, q], p, q) for p, q in baselines])

    from cubical.madmax.flagger import SIGMA_MAD

    def make_antenna_mads(mad_threshold):
        print("make_baseline_mad_plot: plotting antennas", file=log(3))
        # compute per-antenna MAD
        if per_corr:
            ## again, wanted to do this
            # medant = np.ma.median(mad[0,...], axis=(1,2,3))
            ## but not all numpys like it, so being defensive:
            shape1 = mad[0,...].shape
            shape1 = [shape1[0], shape1[1]*shape1[2]*shape1[3]] + list(shape1[4:])
            medant = np.ma.median(mad[0,...].reshape(shape1), axis=1)
        else:
            medant = np.ma.median(mad[0,...], axis=1)
        antnum = np.ma.masked_array(range(n_ant), medant.mask)
        if not medant.mask.all():
            medmed = np.ma.median(medant)
            madmed = np.ma.median(abs(medant-medmed))

            # make plot
            try:
                pylab.plot(antnum, medant, ls='-', color='white')
                pylab.axhline(medmed, color='0.5')
                thresholds = [(1,'green'), (2, 'blue'), (3,'orange')]
                if mad_threshold:
                    thresholds.append((mad_threshold, "red"))
                for thr,color in thresholds:
                    pylab.axhline(medmed+thr*SIGMA_MAD*madmed, color=color, ls=':')
                for p in range(n_ant):
                    if antnum.mask is np.ma.nomask or not antnum.mask[p]:
                        pylab.axvline(antnum[p], color="0.9")
                        color = "black"
                        for thr, col in thresholds:
                            if medant[p] > medmed + thr*SIGMA_MAD*madmed:
                                color = col
                        pylab.text(antnum[p], medant[p], metadata.antenna_name[p], color=color, ha='center', va='center')
                pylab.xticks([])
                pylab.xlabel("antenna")
                pylab.ylabel("MAD residual over baselines")
            except Exception as exc:
                traceback.print_exc()
                print("WARNING: {}: exception {} raised while generating Mad Max antenna-MAD plot".format(
                                        chunk_label, exc), file=log(1,"red"))
                print("Although harmless, this may indicate a problem with the data, or a bug in CubiCal.", file=log(1))
                print("Please see stack trace above, and report if you think this is a bug.", file=log(1))


            if mad_threshold:
                antmask = medant > medmed + mad_threshold*SIGMA_MAD*madmed
                if antmask.any():
                    print("{}: antennas {} have mad residuals, refer to Mad Max plots".format(max_label,
                                                ",".join([metadata.antenna_name[p] for p,fl in enumerate(antmask) if fl])), file=log(0,"red"))
                else:
                    print("{}: no antennas with mad residuals".format(max_label), file=log(1))
                outflags[antmask,:] = True
                outflags[:,antmask] = True
                if per_corr:
                    mad.mask |= antmask[np.newaxis,:,np.newaxis,np.newaxis,np.newaxis]
                    mad.mask |= antmask[np.newaxis,np.newaxis,:,np.newaxis,np.newaxis]
                else:
                    mad.mask |= antmask[np.newaxis, :, np.newaxis]
                    mad.mask |= antmask[np.newaxis, np.newaxis, :]
                return antmask
        return None

    def make_baseline_mads():
        # remake indices, since flagged (masked) baselines may have changed
        if per_corr:
            indices = [(bl,p,q,c1,c2) for bl,p,q in indices_pq for c1 in range(n_cor) for c2 in range(n_cor)]
            mask = [mad.mask[0,p,q,c1,c2] for _,p,q,c1,c2 in indices]
            blmad = np.ma.masked_array([(mad[0,p,q,c1,c2] or 0) for _,p,q,c1,c2 in indices],mask)
        else:
            indices = [(bl,p,q,None,None) for bl,p,q in indices_pq]
            mask = [mad.mask[0,p,q] for _,p,q in indices_pq]
            blmad = np.ma.masked_array([(mad[0,p,q] or 0) for _,p,q in indices_pq],mask)

        bllen = np.ma.masked_array([bl for bl,_,_,_,_ in indices],mask)
        # N,2 array of pq indices for each entry
        blpq  = np.array([[p,q] for bl,p,q,_,_ in indices])

        # for every baseline, compute local MMAD
        from cubical.madmax.flagger import SIGMA_MAD
        print("make_baseline_mad_plot: computing LMMAD", file=log(3))
        lmmad = {}
        for i,(_,p,q,_,_) in enumerate(indices):
            if (p,q) not in lmmad:
                bllen0 = metadata.baseline_length[p,q]
                selection = (bllen>0.8*bllen0)&(bllen<1.2*bllen0)&((p!=blpq[:,0])|(q!=blpq[:,1]))
                selection = selection&~blmad.mask
                if selection.any():
                    med = np.ma.median(blmad[selection&~blmad.mask])
                    if med is not np.ma.masked:
                        lmmad[p,q] = med

        lmmad_threshold = 3
        lmmad_mask = [((p,q) not in lmmad) for _,p,q in indices_pq]
        lmmad_bllen = np.ma.masked_array([bl for bl,_,_ in indices_pq], lmmad_mask)
        lmmad_mad   = np.ma.masked_array([(lmmad.get((p,q), 0) or 0) for _,p,q in indices_pq], lmmad_mask)

        lmmad_ad = np.ma.masked_array([abs((blmad1 or 0) - lmmad.get((p,q), 0)) for (_,p,q,_,_),blmad1 in zip(indices,blmad)], blmad.mask)
        lmmad_madmad = np.ma.median(lmmad_ad)
        print("make_baseline_mad_plot: plotting baselines", file=log(3))


        xlim = [0, 0]
        ylim = [0, 0]
        try:
            wh = ~lmmad_bllen.mask
            pylab.fill_between(lmmad_bllen[wh], lmmad_mad[wh], (lmmad_mad+lmmad_madmad*lmmad_threshold)[wh], color='0.8', step='mid')
            if med_thr is not None:
                if per_corr:
                    for ic1, c1 in enumerate(metadata.feeds.upper()):
                        for ic2, c2 in enumerate(metadata.feeds.upper()):
                            if not medmad.mask[0, ic1, ic2]: 
                                pylab.axhline(medmad[0, ic1, ic2], ls="-", color=colors[ic1][ic2])
                                pylab.text(0, medmad[0, ic1, ic2], "MMAD", color=colors[ic1][ic2],
                                       ha='right', va='center', size='x-small')
                                if med_thr is not None:
                                    pylab.axhline(med_thr[0, ic1, ic2], ls=":", color=colors[ic1][ic2])
                                    pylab.text(0, med_thr[0, ic1, ic2], "threshold", color=colors[ic1][ic2],
                                           ha='right', va='center', size='x-small')
                else:
                    pylab.axhline(medmad[0], ls="-", color="blue")
                    pylab.text(0, medmad[0], "MMAD", color="blue",
                               ha='right', va='center', size='x-small')
                    if med_thr is not None:
                        pylab.axhline(med_thr[0], ls=":", color="black")
                        pylab.text(0, med_thr[0], "threshold", color="black",
                                   ha='right', va='center', size='x-small')

            for p in range(n_ant):
                for q in range(p + 1, n_ant):
                    if not mad.mask[0, p, q].all():
                        uvdist = metadata.baseline_length[p, q]
                        xlim[1] = max(xlim[1], uvdist)
                        if per_corr:
                            for ic1, c1 in enumerate(metadata.feeds.upper()):
                                for ic2, c2 in enumerate(metadata.feeds.upper()):
                                    if not mad.mask[0, p, q, ic1, ic2]:
                                        y = mad[0, p, q, ic1, ic2]
                                        pylab.text(uvdist, y, metadata.baseline_name[p, q],
                                                   color=colors[ic1][ic2],
                                                   horizontalalignment='center', verticalalignment='center')
                                        ylim[0] = min(ylim[0], y)
                                        ylim[1] = max(ylim[1], y)
                        else:
                            y = mad[0, p, q]
                            pylab.text(uvdist, y, metadata.baseline_name[p, q],
                                       horizontalalignment='center', verticalalignment='center')
                            ylim[0] = min(ylim[0], y)
                            ylim[1] = max(ylim[1], y)

            import matplotlib.lines as mlines
            if per_corr:
                handles = []
                for ic1, c1 in enumerate(metadata.feeds):
                    for ic2, c2 in enumerate(metadata.feeds):
                        handles.append(mlines.Line2D([], [], color=colors[ic1][ic2], label="{}{}".format(c1, c2).upper()))
            else:
                handles = [ mlines.Line2D([], [], color="blue", ls="_", label="MMAD"),
                            mlines.Line2D([], [], color="black", label="threshold") ]
            handles.append(mlines.Line2D([], [], color="0.8", label="LMMAD*{:.2f}".format(lmmad_threshold)))
            pylab.legend(handles=handles)
            pylab.xlim(*xlim)
            pylab.ylim(*ylim)
            pylab.xlabel("Baseline, m.")
            pylab.ylabel("MAD residuals")
            pylab.title("{}: MAD residuals".format(max_label))
        except Exception as exc:
            traceback.print_exc()
            print("WARNING: {}: exception {} raised while generating Mad Max baseline-MAD plot".format(
                chunk_label, exc), file=log(1, "red"))
            print("Although harmless, this may indicate a problem with the data, or a bug in CubiCal.", file=log(1))
            print("Please see stack trace above, and report if you think this is a bug.", file=log(1))


    figure = pylab.figure(figsize=(16, 10))
    pylab.subplot(2,2,2)
    make_baseline_mads()

    pylab.subplot(2,2,1)
    make_antenna_mads(antenna_mad_threshold)

    pylab.subplot(2,2,4)
    make_baseline_mads()

    pylab.subplot(2,2,3)
    make_antenna_mads(antenna_mad_threshold)

    print("make_baseline_mad_plot: done", file=log(3))

    return outflags, figure




from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_2x2_inverse(x, xinv, flags, eps, flagbit):
    """
    Given an array x of dimensions (d,t,i,a,2,2), computes the inverse of every 2x2 block. 
    Takes flags of shape (d,t,f,a) into account, and will flag elements if the inverse is 
    too large.

    Args:
        x (np.complex64 or np.complex128):
            Typed memoryview of X array with dimensions (d, ti, fi, a, c, c)
        xinv (np.complex64 or np.complex128):
            Typed memoryview of output inverse array with dimensions
            (d, ti, fi, a, c, c)
        flags (np.uint16_t):
            Typed memoryview of flag array with dimensions (d, t, f, a)
        eps (float):
            Threshold beneath which the denominator is regarded as too small for inversion.
        flagbit (int):
            The bitflag which will be raised if flagging is required.

    Returns:
        int:
            Number of elements flagged
    """

    flag_count = 0

    eps = eps**2

    n_dir = x.shape[0]
    n_tim = x.shape[1]
    n_fre = x.shape[2]
    n_ant = x.shape[3]

    for aa in prange(n_ant):
        for t in range(n_tim):
            for f in range(n_fre):
                for d in range(n_dir):
                    if flags[d,t,f,aa]:

                        xinv[d,t,f,aa,0,0] = 0
                        xinv[d,t,f,aa,1,1] = 0
                        xinv[d,t,f,aa,0,1] = 0
                        xinv[d,t,f,aa,1,0] = 0

                    else:
                        denom = x[d,t,f,aa,0,0] * x[d,t,f,aa,1,1] - \
                                x[d,t,f,aa,0,1] * x[d,t,f,aa,1,0]

                        if (denom*denom.conjugate()).real<=eps:
                            
                            xinv[d,t,f,aa,0,0] = 0
                            xinv[d,t,f,aa,1,1] = 0
                            xinv[d,t,f,aa,0,1] = 0
                            xinv[d,t,f,aa,1,0] = 0

                            flags[d,t,f,aa] = flagbit
                            flag_count += 1

                        else:

                            xinv[d,t,f,aa,0,0] = x[d,t,f,aa,1,1]/denom
                            xinv[d,t,f,aa,1,1] = x[d,t,f,aa,0,0]/denom
                            xinv[d,t,f,aa,0,1] = -1 * x[d,t,f,aa,0,1]/denom
                            xinv[d,t,f,aa,1,0] = -1 * x[d,t,f,aa,1,0]/denom

    return flag_count

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_diag_inverse(x, xinv, flags, eps, flagbit):
    """
    Given an array x of dimensions (d,t,i,a,2,2), computes the inverse of every 2x2 block,
    under the assumption that the off-diagonal entries are zero. Takes flags of shape 
    (d,t,f,a) into account, and will flag elements if the inverse is too large.

    Args:
        x (np.complex64 or np.complex128):
            Typed memoryview of X array with dimensions (d, ti, fi, a, c, c)
        xinv (np.complex64 or np.complex128):
            Typed memoryview of output inverse array with dimensions
            (d, ti, fi, a, c, c)
        flags (np.uint16_t):
            Typed memoryview of flag array with dimensions (d, t, f, a)
        eps (float):
            Threshold beneath which the denominator is regarded as too small for inversion.
        flagbit (int):
            The bitflag which will be raised if flagging is required.

    Returns:
        int:
            Number of elements flagged
    """

    flag_count = 0
    eps = eps**2

    n_dir = x.shape[0]
    n_tim = x.shape[1]
    n_fre = x.shape[2]
    n_ant = x.shape[3]

    for aa in prange(n_ant):
        for t in range(n_tim):
            for f in range(n_fre):
                for d in range(n_dir):
                    
                    xinv[d,t,f,aa,0,1] = xinv[d,t,f,aa,1,0] = 0
                    
                    if flags[d,t,f,aa]:
                        xinv[d,t,f,aa,0,0] = xinv[d,t,f,aa,1,1] = 0
                    
                    else:
                        denom = x[d,t,f,aa,0,0] * x[d,t,f,aa,1,1]
                        if (denom.real**2 + denom.imag**2)<=eps:
                            xinv[d,t,f,aa,0,0] = xinv[d,t,f,aa,1,1] = 0
                            flags[d,t,f,aa] = flagbit
                            flag_count += 1
                        else:
                            xinv[d,t,f,aa,0,0] = 1/x[d,t,f,aa,0,0]
                            xinv[d,t,f,aa,1,1] = 1/x[d,t,f,aa,1,1]

    return flag_count

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_chisq(r, chisq):
    """
    Compute chi-square over correlations, models, and one antenna axis.

    Args:
        r (np.complex64 or np.complex128):
            Array with dimensions (i, t, f , a, a, c, c)
        chisq (np.float64):
            Array with dimensions (t, f, a)
    """
    n_mod = r.shape[0]
    n_tim = r.shape[1]
    n_fre = r.shape[2]
    n_ant = r.shape[3]

    for aa in prange(n_ant):
        for ab in range(n_ant):
            for i in range(n_mod):
                for t in range(n_tim):
                    for f in range(n_fre):
                        for c1 in range(2):
                            for c2 in range(2):
                                chisq[t,f,aa] += r[i,t,f,aa,ab,c1,c2].real**2 + r[i,t,f,aa,ab,c1,c2].imag**2

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_chisq_diag(r, chisq):
    """
    Compute chi-square over diagonal correlations, models, and one antenna axis.

    Args:
        r (np.complex64 or np.complex128):
            Array with dimensions (i, t, f , a, a, c, c)
        chisq (np.float64):
            Array with dimensions (t, f, a)
    """
    n_mod = r.shape[0]
    n_tim = r.shape[1]
    n_fre = r.shape[2]
    n_ant = r.shape[3]

    for aa in prange(n_ant):
        for ab in range(n_ant):
            for i in range(n_mod):
                for t in range(n_tim):
                    for f in range(n_fre):
                        for c in range(2):
                            chisq[t,f,aa] += r[i,t,f,aa,ab,c,c].real**2 + r[i,t,f,aa,ab,c,c].imag**2

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_chisq_offdiag(r, chisq):
    """
    Compute chi-square over off-diagonal correlations, models, and one antenna axis.

    Args:
        r (np.complex64 or np.complex128):
            Array with dimensions (i, t, f , a, a, c, c)
        chisq (np.float64):
            Array with dimensions (t, f, a)
    """
    n_mod = r.shape[0]
    n_tim = r.shape[1]
    n_fre = r.shape[2]
    n_ant = r.shape[3]

    for aa in prange(n_ant):
        for ab in range(n_ant):
            for i in range(n_mod):
                for t in range(n_tim):
                    for f in range(n_fre):
                        for c in range(2):
                            chisq[t,f,aa] += r[i,t,f,aa,ab,c,1-c].real**2 + r[i,t,f,aa,ab,c,1-c].imag**2



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

# This module has been adapted from the DDFacet package,
# (c) Cyril Tasse et al., see http://github.com/saopicc/DDFacet

HEADER = '\033[95m'
OKBLUE = '\033[94m'
OKGREEN = '\033[92m'
WARNING = '\033[93m'
FAIL = '\033[91m'
ENDC = '\033[0m'
bold='\033[1m'
nobold='\033[0m'
Separator="================================%s=================================="
silent=0

_color_dict = dict(red=FAIL, green=OKGREEN, yellow=WARNING, blue=OKBLUE,
                   white="")

def disableColors():
    global silent
    silent = 1
    
def Str(strin0,col="red",Bold=True):
    if silent==1: return strin0
    strin=str(strin0)

    ss = _color_dict.get(col)
    if ss is None:
        raise ValueError("unknown color '{}'".format(col))

    ss="%s%s%s"%(ss,strin,ENDC)
    if Bold: ss="%s%s%s"%(bold,ss,nobold)
    return ss

def Sep(strin=None,D=1):
    if D!=1:
        return Str(Separator%("="*len(strin)))
    else:
        return Str(Separator%(strin))

def Title(strin,Big=False):
    print()
    print()
    if Big: print(Sep(strin,D=0))
    print(Sep(strin))
    if Big: print(Sep(strin,D=0))
    print()

def disable():
    HEADER = ''
    OKBLUE = ''
    OKGREEN = ''
    WARNING = ''
    FAIL = ''
    ENDC = ''
